import os
import tempfile
import giturlparse
from typing import Iterator, Optional, TYPE_CHECKING
from contextlib import contextmanager

from dlt.common.exceptions import MissingDependencyException
from dlt.common.storages import FileStorage
from dlt.common.utils import uniq_id
from dlt.common.typing import Any


# NOTE: never import git module directly as it performs a check if the git command is available and raises ImportError
if TYPE_CHECKING:
    from git import Repo
else:
    Repo = Any


def _import_git() -> None:
    try:
        import git
    except ModuleNotFoundError:
        raise MissingDependencyException(
            "git repository helpers",
            ["gitpython>=3.1.29"],
            "Install PythonGit to work with git repositories.",
        )


@contextmanager
def git_custom_key_command(private_key: Optional[str]) -> Iterator[str]:
    if private_key:
        key_file = tempfile.mktemp(prefix=uniq_id())
        with open(key_file, "w", encoding="utf-8") as f:
            f.write(private_key)
        try:
            # permissions so SSH does not complain
            os.chmod(key_file, 0o600)
            yield 'ssh -o "StrictHostKeyChecking accept-new" -i "%s"' % key_file.replace(
                "\\", "\\\\"
            )
        finally:
            os.remove(key_file)
    else:
        yield 'ssh -o "StrictHostKeyChecking accept-new"'


def is_clean_and_synced(repo: Repo) -> bool:
    """Checks if repo is clean and synced with origin"""
    # get branch status
    status: str = repo.git.status("--short", "--branch")
    # we expect first status line ## main...origin/main
    status_lines = status.splitlines()
    first_line = status_lines[0]
    # we expect first status line is not ## main...origin/main [ahead 1]
    return len(status_lines) == 1 and first_line.startswith("##") and not first_line.endswith("]")


def is_dirty(repo: Repo) -> bool:
    status: str = repo.git.status("--short")
    return len(status.strip()) > 0


def get_default_branch(repo: Repo) -> str:
    origin = repo.remotes.origin
    # Get the remote's HEAD reference (default branch)
    default_ref = origin.refs.HEAD  # symbolic ref like 'origin/main'
    return default_ref.reference.name.split("/")[-1]  # extract branch


def ensure_remote_head(
    repo_path: str,
    branch: Optional[str] = None,
    with_git_command: Optional[str] = None,
    path: Optional[str] = None,
) -> None:
    """Updates repository from origin and ensures it's clean and synced.

    Uses sparse checkout when path is specified, fetching only the specified
    directory instead of the entire repository tree.

    Args:
        repo_path: Local path to the git repository.
        branch: Branch to checkout. Defaults to repository's default branch.
        with_git_command: Custom GIT_SSH_COMMAND for authentication.
        path: Directory path for sparse checkout. When set, only this path
            is checked out, reducing download size and time.

    Raises:
        RepositoryDirtyError: If repository has uncommitted changes or is
            not synced with origin.
    """
    from git import Repo, RepositoryDirtyError

    # update remotes and check if heads are same. ignores locally modified files
    with Repo(repo_path) as repo:
        # use custom environment if specified
        with repo.git.custom_environment(GIT_SSH_COMMAND=with_git_command):
            # if path is set, use sparse checkout
            if path is not None:
                # assume that sparse checkout was enabled when cloning
                repo.git.sparse_checkout("set", path)
            # checkout branch before fetching
            repo.git.checkout(branch or get_default_branch(repo))
            # update origin
            repo.remote().pull()
            if not is_clean_and_synced(repo):
                status: str = repo.git.status("--short", "--branch")
                raise RepositoryDirtyError(repo, status)


def clone_repo(
    repository_url: str,
    clone_path: str,
    branch: Optional[str] = None,
    with_git_command: Optional[str] = None,
    path: Optional[str] = None,
) -> Repo:
    _import_git()

    from git import Repo

    if path is not None:
        # set of options that prevents downloading all blobs
        multi_options = [
            "--depth=1",
            "--filter=blob:none",
            "--no-checkout",
        ]
    else:
        multi_options = None
    repo = Repo.clone_from(
        repository_url,
        clone_path,
        env=dict(GIT_SSH_COMMAND=with_git_command),
        multi_options=multi_options,
    )
    # set up sparse mode to checkout paths on demand
    if path is not None:
        repo.git.sparse_checkout("init", "--cone")
        repo.git.sparse_checkout("set", path)
    if branch:
        repo.git.checkout(branch)
    elif path is not None:
        repo.git.checkout()
    return repo


def force_clone_repo(
    repo_url: str,
    repo_storage: FileStorage,
    repo_name: str,
    branch: Optional[str] = None,
    with_git_command: Optional[str] = None,
    path: Optional[str] = None,
) -> None:
    """Deletes existing repository and performs fresh clone.

    Removes repo_storage.root/repo_name if it exists, then clones from
    repo_url. Uses sparse checkout when path is specified to download only
    the specified directory, reducing clone time and disk usage.

    Args:
        repo_url: Git repository URL to clone from.
        repo_storage: FileStorage instance managing the clone destination.
        repo_name: Directory name for the cloned repository.
        branch: Branch to checkout after cloning.
        with_git_command: Custom GIT_SSH_COMMAND for authentication.
        path: Directory path for sparse checkout. When set, only this path
            is cloned using --filter=blob:none and --depth=1.
    """
    try:
        # delete repo folder
        if repo_storage.has_folder(repo_name):
            repo_storage.delete_folder(repo_name, recursively=True, delete_ro=True)
        clone_repo(
            repo_url,
            repo_storage.make_full_path(repo_name),
            branch=branch,
            with_git_command=with_git_command,
            path=path,
        ).close()
    except Exception:
        # delete folder so we start clean next time
        if repo_storage.has_folder(repo_name):
            repo_storage.delete_folder(repo_name, recursively=True, delete_ro=True)
        raise


def get_fresh_repo_files(
    repo_location: str,
    working_dir: str = None,
    branch: Optional[str] = None,
    with_git_command: Optional[str] = None,
    path: Optional[str] = None,
) -> FileStorage:
    """Returns FileStorage with up-to-date repository files.

    If repo_location is a local directory, returns storage pointing to it.
    If it's a git URL, clones or updates the repository in working_dir/repo_name.
    Supports sparse checkout to fetch only a specific directory path.

    Args:
        repo_location: Local directory path or git repository URL.
        working_dir: Directory where repository will be cloned if repo_location is URL.
        branch: Branch to checkout.
        with_git_command: Custom GIT_SSH_COMMAND for authentication.
        path: Directory path for sparse checkout. Downloads only this directory,
            improving performance for large repositories.

    Returns:
        FileStorage instance pointing to repository files (or specified path within).
    """
    from git import GitError

    url = giturlparse.parse(repo_location, check_domain=False)
    if not url.valid:
        # repo is a directory so jus return storage
        return FileStorage(repo_location, makedirs=False)
    else:
        # clone or update repo
        repo_name = url.name
        repo_path = os.path.join(working_dir, repo_name)
        try:
            ensure_remote_head(
                repo_path, branch=branch, with_git_command=with_git_command, path=path
            )
        except GitError:
            force_clone_repo(
                repo_location,
                FileStorage(working_dir, makedirs=True),
                repo_name,
                branch=branch,
                with_git_command=with_git_command,
                path=path,
            )
        return FileStorage(repo_path)


def get_repo(path: str) -> Repo:
    _import_git()

    from git import Repo

    # if GIT_CEILING_DIRECTORIES is set then do not look up for repositories in parent dirs
    search_parent_directories: bool = True
    if os.getenv("GIT_CEILING_DIRECTORIES"):
        search_parent_directories = False

    repo = Repo(path, search_parent_directories=search_parent_directories)
    return repo


def get_origin(repo: Repo) -> str:
    return repo.remote().url



def join_url(base_url: str, path: str) -> str:
    if base_url is None:
        raise ValueError("Base URL must be provided or set to an empty string.")

    if base_url == "":
        return path

    if path == "":
        return base_url

    # Normalize the base URL
    base_url = base_url.rstrip("/")
    if not base_url.endswith("/"):
        base_url += "/"

    return base_url + path.lstrip("/")



from types import TracebackType
from typing import (
    List,
    Any,
    cast,
    Union,
    Tuple,
    Iterable,
    Type,
    Optional,
    Sequence,
    TYPE_CHECKING,
)

import lancedb
import lancedb.table
import pyarrow as pa
from lancedb.embeddings import EmbeddingFunctionRegistry, TextEmbeddingFunction
from lancedb.query import LanceQueryBuilder
from pyarrow import Array, ChunkedArray

from dlt.common import json, pendulum, logger
from dlt.common.libs.numpy import numpy
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.destination.exceptions import (
    DestinationUndefinedEntity,
    DestinationTerminalException,
)
from dlt.common.destination.client import (
    JobClientBase,
    PreparedTableSchema,
    WithStateSync,
    StorageSchemaInfo,
    StateInfo,
    LoadJob,
    FollowupJobRequest,
)
from dlt.common.schema import Schema, TSchemaTables
from dlt.common.schema.typing import (
    C_DLT_LOADS_TABLE_LOAD_ID,
    TTableSchemaColumns,
    TColumnSchema,
    TTableSchema,
)
from dlt.common.schema.utils import (
    get_columns_names_with_prop,
    is_nested_table,
)
from dlt.common.storages import FileStorage, LoadJobInfo, ParsedLoadJobFileName
from dlt.destinations.impl.lancedb.configuration import (
    LanceDBClientConfiguration,
)
from dlt.destinations.impl.lancedb.exceptions import (
    lancedb_error,
)
from dlt.destinations.impl.lancedb.jobs import LanceDBLoadJob, LanceDBRemoveOrphansJob
from dlt.destinations.impl.lancedb.lancedb_adapter import (
    VECTORIZE_HINT,
    NO_REMOVE_ORPHANS_HINT,
)
from dlt.destinations.impl.lancedb.schema import (
    make_arrow_field_schema,
    make_arrow_table_schema,
    TArrowSchema,
    NULL_SCHEMA,
    TArrowField,
)
from dlt.destinations.impl.lancedb.utils import (
    set_non_standard_providers_environment_variables,
    write_records,
)
from dlt.destinations.job_impl import ReferenceFollowupJobRequest
from dlt.destinations.sql_jobs import SqlMergeFollowupJob
from dlt.destinations.type_mapping import TypeMapperImpl
from dlt.destinations.sql_client import SqlClientBase, WithSqlClient

if TYPE_CHECKING:
    NDArray = numpy.ndarray[Any, Any]
else:
    NDArray = numpy.ndarray


class LanceDBClient(JobClientBase, WithStateSync, WithSqlClient):
    """LanceDB destination handler."""

    model_func: TextEmbeddingFunction
    """The embedder callback used for each chunk."""
    dataset_name: str

    def __init__(
        self,
        schema: Schema,
        config: LanceDBClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        super().__init__(schema, config, capabilities)
        self.registry = EmbeddingFunctionRegistry.get_instance()
        self.config: LanceDBClientConfiguration = config
        self.db_client = self.config.credentials.get_conn()
        self.type_mapper = self.capabilities.get_type_mapper()
        self.sentinel_table_name = config.sentinel_table_name
        self.dataset_name = self.config.normalize_dataset_name(self.schema)
        self._sql_client: SqlClientBase[Any] = None

        embedding_model_provider = self.config.embedding_model_provider
        embedding_model_host = self.config.embedding_model_provider_host

        # LanceDB doesn't provide a standardized way to set API keys across providers.
        # Some use ENV variables and others allow passing api key as an argument.
        # To account for this, we set provider environment variable as well.
        set_non_standard_providers_environment_variables(
            embedding_model_provider,
            self.config.credentials.embedding_model_provider_api_key,
        )

        self.model_func = self.registry.get(embedding_model_provider).create(
            name=self.config.embedding_model,
            max_retries=self.config.options.max_retries,
            # actually the model func doesnt need the api-key!
            **({"host": embedding_model_host} if embedding_model_host else {}),
        )

    @property
    def sql_client_class(self) -> Type[SqlClientBase[Any]]:
        from dlt.destinations.impl.lancedb.sql_client import LanceDBSQLClient

        return LanceDBSQLClient

    @property
    def sql_client(self) -> SqlClientBase[Any]:
        # inner import because `LanceDBSQLClient` depends on `duckdb` and is optional
        from dlt.destinations.impl.lancedb.sql_client import LanceDBSQLClient

        if not self._sql_client:
            self._sql_client = LanceDBSQLClient(self)
        return self._sql_client

    @sql_client.setter
    def sql_client(self, client: SqlClientBase[Any]) -> None:
        self._sql_client = client

    @property
    def sentinel_table(self) -> str:
        return self.make_qualified_table_name(self.sentinel_table_name)

    def make_qualified_table_name(self, table_name: str) -> str:
        return (
            f"{self.dataset_name}{self.config.dataset_separator}{table_name}"
            if self.dataset_name
            else table_name
        )

    def get_table_schema(self, table_name: str) -> TArrowSchema:
        schema_table: "lancedb.table.Table" = self.db_client.open_table(table_name)
        schema_table.checkout_latest()
        schema = schema_table.schema
        return cast(
            TArrowSchema,
            schema,
        )

    @lancedb_error
    def create_table(
        self, table_name: str, schema: TArrowSchema, mode: str = "create"
    ) -> "lancedb.table.Table":
        """Create a LanceDB Table from the provided LanceModel or PyArrow schema.

        Args:
            schema: The table schema to create.
            table_name: The name of the table to create.
            mode (str): The mode to use when creating the table. Can be either "create" or "overwrite".
                By default, if the table already exists, an exception is raised.
                If you want to overwrite the table, use mode="overwrite".
        """
        return self.db_client.create_table(table_name, schema=schema, mode=mode)

    def drop_tables(self, *tables: str, delete_schema: bool = True) -> None:
        """Drop multiple LanceDB tables.

        Args:
            table_names: The names of the tables to drop.
        """
        if not tables:
            return

        for table_name in tables:
            if table_name in self.db_client.table_names():
                self.db_client.drop_table(table_name)

    def delete_table(self, table_name: str) -> None:
        """Delete a LanceDB table.

        Args:
            table_name: The name of the table to delete.
        """
        self.db_client.drop_table(table_name)

    def query_table(
        self,
        table_name: str,
        query: Union[List[Any], NDArray, Array, ChunkedArray, str, Tuple[Any], None] = None,
    ) -> LanceQueryBuilder:
        """Query a LanceDB table.

        Args:
            table_name: The name of the table to query.
            query: The targeted vector to search for.

        Returns:
            A LanceDB query builder.
        """
        query_table: "lancedb.table.Table" = self.db_client.open_table(table_name)
        query_table.checkout_latest()
        return query_table.search(query=query)

    @lancedb_error
    def _get_table_names(self) -> List[str]:
        """Return all tables in the dataset, excluding the sentinel table."""
        if self.dataset_name:
            prefix = f"{self.dataset_name}{self.config.dataset_separator}"
            table_names = [
                table_name
                for table_name in self.db_client.table_names()
                if table_name.startswith(prefix)
            ]
        else:
            table_names = self.db_client.table_names()

        return [table_name for table_name in table_names if table_name != self.sentinel_table]

    @lancedb_error
    def drop_storage(self) -> None:
        """Drop the dataset from the LanceDB instance.

        Deletes all tables in the dataset and all data, as well as sentinel table associated with them.

        If the dataset name wasn't provided, it deletes all the tables in the current schema.
        """
        for table_name in self._get_table_names():
            self.db_client.drop_table(table_name)

        self._delete_sentinel_table()

    @lancedb_error
    def initialize_storage(self, truncate_tables: Iterable[str] = None) -> None:
        if not self.is_storage_initialized():
            self._create_sentinel_table()
        elif truncate_tables:
            for table_name in truncate_tables:
                fq_table_name = self.make_qualified_table_name(table_name)
                if not self.table_exists(fq_table_name):
                    continue
                schema = self.get_table_schema(fq_table_name)
                self.db_client.drop_table(fq_table_name)
                self.create_table(
                    table_name=fq_table_name,
                    schema=schema,
                )

    @lancedb_error
    def is_storage_initialized(self) -> bool:
        return self.table_exists(self.sentinel_table)

    def verify_schema(
        self, only_tables: Iterable[str] = None, new_jobs: Iterable[ParsedLoadJobFileName] = None
    ) -> List[PreparedTableSchema]:
        loaded_tables = super().verify_schema(only_tables, new_jobs)

        # Verify LanceDB-specific requirements for root tables
        for load_table in loaded_tables:
            # Skip nested tables as they inherit behavior from parent tables
            if is_nested_table(load_table):
                continue

            # Check if this table has orphan removal enabled (either explicitly or via merge strategy)
            has_orphan_removal = not load_table.get(NO_REMOVE_ORPHANS_HINT)
            merge_keys = get_columns_names_with_prop(load_table, "merge_key")
            uses_merge_strategy = load_table.get("write_disposition", "") == "merge"

            # Validate merge key constraints when orphan removal is enabled
            if has_orphan_removal and len(merge_keys) > 1:
                raise DestinationTerminalException(
                    "Multiple merge keys are not supported when LanceDB orphan removal is"
                    f" enabled: {merge_keys}"
                )

            # Check if _dlt_load_id column is required but not present
            requires_dlt_ids = has_orphan_removal and uses_merge_strategy
            if requires_dlt_ids and "_dlt_load_id" not in load_table["columns"].keys():
                raise DestinationTerminalException(
                    "The `_dlt_load_id` column is required for tables with orphan removal or merge"
                    " keys. Enable this by setting"
                    " `NORMALIZE__PARQUET_NORMALIZER__ADD_DLT_LOAD_ID=TRUE` or an equivalent in"
                    " config.toml."
                )

        return loaded_tables

    def _create_sentinel_table(self) -> "lancedb.table.Table":
        """Create an empty table to indicate that the storage is initialized."""
        return self.create_table(schema=NULL_SCHEMA, table_name=self.sentinel_table)

    def _delete_sentinel_table(self) -> None:
        """Delete the sentinel table."""
        self.db_client.drop_table(self.sentinel_table)

    @lancedb_error
    def update_stored_schema(
        self,
        only_tables: Iterable[str] = None,
        expected_update: TSchemaTables = None,
    ) -> Optional[TSchemaTables]:
        applied_update = super().update_stored_schema(only_tables, expected_update)
        try:
            schema_info = self.get_stored_schema_by_hash(self.schema.stored_version_hash)
        except DestinationUndefinedEntity:
            schema_info = None

        if schema_info is None:
            logger.info(
                f"Schema with hash {self.schema.stored_version_hash} "
                "not found in the storage. upgrading"
            )
            # TODO: return a real updated table schema (like in SQL job client)
            self._execute_schema_update(only_tables)
        else:
            logger.debug(
                f"Schema with hash {self.schema.stored_version_hash} "
                f"inserted at {schema_info.inserted_at} found "
                "in storage, no upgrade required"
            )
        # we assume that expected_update == applied_update so table schemas in dest were not
        # externally changed
        return applied_update

    def get_storage_table(self, table_name: str) -> Tuple[bool, TTableSchemaColumns]:
        table_schema: TTableSchemaColumns = {}

        try:
            fq_table_name = self.make_qualified_table_name(table_name)

            table: "lancedb.table.Table" = self.db_client.open_table(fq_table_name)
            table.checkout_latest()
            arrow_schema: TArrowSchema = table.schema
        except ValueError:
            return False, table_schema

        field: TArrowField
        for field in arrow_schema:
            name = field.name
            table_schema[name] = {
                "name": name,
                **self.type_mapper.from_destination_type(field.type, None, None),
            }
        return True, table_schema

    def get_storage_tables(
        self, table_names: Iterable[str]
    ) -> Iterable[Tuple[bool, TTableSchemaColumns]]:
        for table_name in table_names:
            # mypy fails to resolve table_schema; ty succeeds
            table_exists, table_schema = self.get_storage_table(table_name)
            yield table_name, table_schema  # type: ignore[misc]

    @lancedb_error
    def extend_lancedb_table_schema(self, table_name: str, field_schemas: List[pa.Field]) -> None:
        """Extend LanceDB table schema with empty columns.

        Args:
        table_name: The name of the table to create the fields on.
        field_schemas: The list of PyArrow Fields to create in the target LanceDB table.
        """
        table: "lancedb.table.Table" = self.db_client.open_table(table_name)
        table.checkout_latest()

        # will be fillde with null values and column is nullable by default
        table.add_columns(field_schemas)

        # TODO: Update method below doesn't work for bulk NULL assignments, raise with LanceDB developers.
        # table.update(values={field.name: None})

    def _execute_schema_update(self, only_tables: Iterable[str]) -> None:
        for table_name in only_tables or self.schema.tables:
            exists, existing_columns = self.get_storage_table(table_name)
            new_columns: List[TColumnSchema] = self.schema.get_new_table_columns(
                table_name,
                existing_columns,
                self.capabilities.generates_case_sensitive_identifiers(),
            )
            logger.info(f"Found {len(new_columns)} updates for {table_name} in {self.schema.name}")
            if new_columns:
                if exists:
                    field_schemas: List[TArrowField] = [
                        make_arrow_field_schema(column["name"], column, self.type_mapper)
                        for column in new_columns
                    ]
                    fq_table_name = self.make_qualified_table_name(table_name)
                    self.extend_lancedb_table_schema(fq_table_name, field_schemas)
                else:
                    if table_name not in self.schema.dlt_table_names():
                        embedding_fields = get_columns_names_with_prop(
                            self.schema.get_table(table_name=table_name), VECTORIZE_HINT
                        )
                        vector_field_name = self.config.vector_field_name
                        embedding_model_func = self.model_func
                        embedding_model_dimensions = self.config.embedding_model_dimensions
                    else:
                        embedding_fields = None
                        vector_field_name = None
                        embedding_model_func = None
                        embedding_model_dimensions = None

                    table_schema: TArrowSchema = make_arrow_table_schema(
                        table_name,
                        schema=self.schema,
                        type_mapper=self.type_mapper,
                        embedding_fields=embedding_fields,
                        embedding_model_func=embedding_model_func,
                        embedding_model_dimensions=embedding_model_dimensions,
                        vector_field_name=vector_field_name,
                    )
                    fq_table_name = self.make_qualified_table_name(table_name)
                    self.create_table(fq_table_name, table_schema)

        self.update_schema_in_storage()

    @lancedb_error
    def update_schema_in_storage(self) -> None:
        records = [
            {
                self.schema.naming.normalize_identifier("version"): self.schema.version,
                self.schema.naming.normalize_identifier(
                    "engine_version"
                ): self.schema.ENGINE_VERSION,
                self.schema.naming.normalize_identifier("inserted_at"): pendulum.now(),
                self.schema.naming.normalize_identifier("schema_name"): self.schema.name,
                self.schema.naming.normalize_identifier(
                    "version_hash"
                ): self.schema.stored_version_hash,
                self.schema.naming.normalize_identifier("schema"): json.dumps(
                    self.schema.to_dict()
                ),
            }
        ]
        fq_version_table_name = self.make_qualified_table_name(self.schema.version_table_name)
        write_disposition = self.schema.get_table(self.schema.version_table_name).get(
            "write_disposition"
        )

        write_records(
            records,
            db_client=self.db_client,
            vector_field_name=self.config.vector_field_name,
            table_name=fq_version_table_name,
            write_disposition=write_disposition,
        )

    @lancedb_error
    def get_stored_state(self, pipeline_name: str) -> Optional[StateInfo]:
        """Retrieves the latest completed state for a pipeline."""
        fq_state_table_name = self.make_qualified_table_name(self.schema.state_table_name)
        fq_loads_table_name = self.make_qualified_table_name(self.schema.loads_table_name)

        state_table_: "lancedb.table.Table" = self.db_client.open_table(fq_state_table_name)
        state_table_.checkout_latest()

        loads_table_: "lancedb.table.Table" = self.db_client.open_table(fq_loads_table_name)
        loads_table_.checkout_latest()

        # normalize property names
        p_load_id = self.schema.naming.normalize_identifier(C_DLT_LOADS_TABLE_LOAD_ID)
        p_dlt_load_id = self.schema.naming.normalize_identifier(
            self.schema.data_item_normalizer.c_dlt_load_id  # type: ignore[attr-defined]
        )
        p_pipeline_name = self.schema.naming.normalize_identifier("pipeline_name")
        p_status = self.schema.naming.normalize_identifier("status")
        p_version = self.schema.naming.normalize_identifier("version")
        p_engine_version = self.schema.naming.normalize_identifier("engine_version")
        p_state = self.schema.naming.normalize_identifier("state")
        p_created_at = self.schema.naming.normalize_identifier("created_at")
        p_version_hash = self.schema.naming.normalize_identifier("version_hash")

        # Read the tables into memory as Arrow tables, with pushdown predicates, so we pull as little
        # data into memory as possible.
        state_table = (
            state_table_.search()
            .where(f"`{p_pipeline_name}` = '{pipeline_name}'", prefilter=True)
            .to_arrow()
        )
        loads_table = loads_table_.search().where(f"`{p_status}` = 0", prefilter=True).to_arrow()

        # Join arrow tables in-memory.
        joined_table: pa.Table = state_table.join(
            loads_table, keys=p_dlt_load_id, right_keys=p_load_id, join_type="inner"
        ).sort_by([(p_dlt_load_id, "descending")])

        if joined_table.num_rows == 0:
            return None

        state = joined_table.take([0]).to_pylist()[0]
        return StateInfo(
            version=state[p_version],
            engine_version=state[p_engine_version],
            pipeline_name=state[p_pipeline_name],
            state=state[p_state],
            created_at=pendulum.instance(state[p_created_at]),
            version_hash=state[p_version_hash],
            _dlt_load_id=state[p_dlt_load_id],
        )

    @lancedb_error
    def get_stored_schema_by_hash(self, schema_hash: str) -> Optional[StorageSchemaInfo]:
        fq_version_table_name = self.make_qualified_table_name(self.schema.version_table_name)

        version_table: "lancedb.table.Table" = self.db_client.open_table(fq_version_table_name)
        version_table.checkout_latest()
        p_version_hash = self.schema.naming.normalize_identifier("version_hash")
        p_inserted_at = self.schema.naming.normalize_identifier("inserted_at")
        p_schema_name = self.schema.naming.normalize_identifier("schema_name")
        p_version = self.schema.naming.normalize_identifier("version")
        p_engine_version = self.schema.naming.normalize_identifier("engine_version")
        p_schema = self.schema.naming.normalize_identifier("schema")

        try:
            schemas = (
                version_table.search().where(
                    f'`{p_version_hash}` = "{schema_hash}"', prefilter=True
                )
            ).to_list()

            most_recent_schema = sorted(schemas, key=lambda x: x[p_inserted_at], reverse=True)[0]
            return StorageSchemaInfo(
                version_hash=most_recent_schema[p_version_hash],
                schema_name=most_recent_schema[p_schema_name],
                version=most_recent_schema[p_version],
                engine_version=most_recent_schema[p_engine_version],
                inserted_at=most_recent_schema[p_inserted_at],
                schema=most_recent_schema[p_schema],
            )
        except IndexError:
            return None

    @lancedb_error
    def get_stored_schema(self, schema_name: str = None) -> Optional[StorageSchemaInfo]:
        """Retrieves newest schema from destination storage."""
        fq_version_table_name = self.make_qualified_table_name(self.schema.version_table_name)
        if fq_version_table_name not in self.db_client.table_names():
            return None

        version_table: "lancedb.table.Table" = self.db_client.open_table(fq_version_table_name)
        version_table.checkout_latest()
        p_version_hash = self.schema.naming.normalize_identifier("version_hash")
        p_inserted_at = self.schema.naming.normalize_identifier("inserted_at")
        p_schema_name = self.schema.naming.normalize_identifier("schema_name")
        p_version = self.schema.naming.normalize_identifier("version")
        p_engine_version = self.schema.naming.normalize_identifier("engine_version")
        p_schema = self.schema.naming.normalize_identifier("schema")

        try:
            query = version_table.search()
            if schema_name:
                query = query.where(f'`{p_schema_name}` = "{schema_name}"', prefilter=True)
            schemas = query.to_list()

            most_recent_schema = sorted(schemas, key=lambda x: x[p_inserted_at], reverse=True)[0]
            return StorageSchemaInfo(
                version_hash=most_recent_schema[p_version_hash],
                schema_name=most_recent_schema[p_schema_name],
                version=most_recent_schema[p_version],
                engine_version=most_recent_schema[p_engine_version],
                inserted_at=most_recent_schema[p_inserted_at],
                schema=most_recent_schema[p_schema],
            )
        except IndexError:
            return None

    def __exit__(
        self,
        exc_type: Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        pass

    def __enter__(self) -> "LanceDBClient":
        return self

    @lancedb_error
    def complete_load(self, load_id: str) -> None:
        records = [
            {
                self.schema.naming.normalize_identifier(C_DLT_LOADS_TABLE_LOAD_ID): load_id,
                self.schema.naming.normalize_identifier("schema_name"): self.schema.name,
                self.schema.naming.normalize_identifier("status"): 0,
                self.schema.naming.normalize_identifier("inserted_at"): pendulum.now(),
                self.schema.naming.normalize_identifier(
                    "schema_version_hash"
                ): self.schema.version_hash,
            }
        ]
        fq_loads_table_name = self.make_qualified_table_name(self.schema.loads_table_name)
        write_disposition = self.schema.get_table(self.schema.loads_table_name).get(
            "write_disposition"
        )
        write_records(
            records,
            db_client=self.db_client,
            vector_field_name=self.config.vector_field_name,
            table_name=fq_loads_table_name,
            write_disposition=write_disposition,
        )

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        if ReferenceFollowupJobRequest.is_reference_job(file_path):
            return LanceDBRemoveOrphansJob(file_path)
        else:
            return LanceDBLoadJob(file_path, table)

    def create_table_chain_completed_followup_jobs(
        self,
        table_chain: Sequence[TTableSchema],
        completed_table_chain_jobs: Optional[Sequence[LoadJobInfo]] = None,
    ) -> List[FollowupJobRequest]:
        jobs = super().create_table_chain_completed_followup_jobs(
            table_chain, completed_table_chain_jobs  # type: ignore[arg-type]
        )
        # Orphan removal is only supported for upsert strategy because we need a deterministic key hash.
        first_table_in_chain = table_chain[0]
        if first_table_in_chain.get(
            "write_disposition"
        ) == "merge" and not first_table_in_chain.get(NO_REMOVE_ORPHANS_HINT):
            all_job_paths_ordered = [
                job.file_path
                for table in table_chain
                for job in completed_table_chain_jobs
                if job.job_file_info.table_name == table.get("name")
            ]
            root_table_file_name = FileStorage.get_file_name_from_file_path(
                all_job_paths_ordered[0]
            )
            jobs.append(ReferenceFollowupJobRequest(root_table_file_name, all_job_paths_ordered))
        return jobs

    def table_exists(self, table_name: str) -> bool:
        return table_name in self.db_client.table_names()



from dlt.common.configuration.plugins import SupportsCliCommand

from dlt._workspace.cli.exceptions import CliCommandException

DEFAULT_VERIFIED_SOURCES_REPO = "https://github.com/dlt-hub/verified-sources.git"


__all__ = [
    "SupportsCliCommand",
    "CliCommandException",
    "DEFAULT_VERIFIED_SOURCES_REPO",
]



import os
import sys
from pathlib import Path
from types import ModuleType
from typing import Any, Tuple
from unittest.mock import patch
from importlib import import_module

from dlt.common import logger
from dlt.common.exceptions import DltException
from dlt.common.typing import DictStrAny
from dlt.common.reflection.ref import import_module_with_missing

from dlt.extract import DltSource
from dlt.extract.pipe_iterator import ManagedPipeIterator


def patch__init__(self: Any, *args: Any, **kwargs: Any) -> None:
    raise PipelineIsRunning(self, args, kwargs)


def import_script_module(
    module_path: str, script_relative_path: str, ignore_missing_imports: bool = False
) -> ModuleType:
    """Loads a module in `script_relative_path` by splitting it into a script module (file part) and package (folders).  `module_path` is added to sys.path
    Optionally, missing imports will be ignored by importing a dummy module instead.
    """
    if os.path.isabs(script_relative_path):
        raise ValueError(script_relative_path, f"Not relative path to `{module_path}`")

    module, _ = os.path.splitext(script_relative_path)
    module = ".".join(Path(module).parts)

    # add path to module search
    sys_path: str = None
    if module_path not in sys.path:
        sys_path = module_path
        # path must be first so we always load our module of
        sys.path.insert(0, sys_path)
    try:
        logger.info(f"Importing pipeline script from module `{module}` with path `{module_path}`")
        if ignore_missing_imports:
            return import_module_with_missing(module)
        else:
            return import_module(module)

    finally:
        # remove script module path
        if sys_path:
            sys.path.remove(sys_path)


def import_pipeline_script(
    module_path: str, script_relative_path: str, ignore_missing_imports: bool = False
) -> ModuleType:
    # patch entry points to pipeline, sources and resources to prevent pipeline from running
    with (
        patch.object(DltSource, "__init__", patch__init__),
        patch.object(ManagedPipeIterator, "__init__", patch__init__),
    ):
        return import_script_module(
            module_path, script_relative_path, ignore_missing_imports=ignore_missing_imports
        )


class PipelineIsRunning(DltException):
    def __init__(self, obj: object, args: Tuple[str, ...], kwargs: DictStrAny) -> None:
        if isinstance(obj, ManagedPipeIterator):
            msg = "The pipeline script runs the pipeline on import."
        else:
            msg = f"The pipeline script instantiates {obj.__class__.__name__} on import."
        msg += (
            " Initialize sources, and run the pipeline in script entry point:\n"
            "if __name__ == 'main':\n  ..."
        )
        super().__init__(
            msg,
            obj,
            args,
            kwargs,
        )



from typing import Any, AnyStr, Collection, Dict, List, Sequence, Optional, Type
from dlt.common.typing import TypedDict


class ExceptionTrace(TypedDict, total=False):
    """Exception trace. NOTE: we intend to change it with an extended line by line trace with code snippets"""

    message: str
    exception_type: str
    docstring: str
    stack_trace: List[str]
    is_terminal: bool
    """Says if exception is terminal if happened to a job during load step"""
    exception_attrs: Dict[str, Any]
    """Public attributes of an exception deriving from DltException (not starting with _)"""
    load_id: str
    """Load id if found in exception attributes"""
    pipeline_name: str
    """Pipeline name if found in exception attributes or in the active pipeline (Container)"""
    source_name: str
    """Source name if found in exception attributes or in Container"""
    resource_name: str
    """Resource name if found in exception attributes"""
    job_id: str
    """Job id if found in exception attributes"""


class DltException(Exception):
    def __reduce__(self) -> Any:
        """Enables exceptions with parametrized constructor to be pickled"""
        return type(self).__new__, (type(self), *self.args), self.__dict__

    def attrs(self) -> Dict[str, Any]:
        """Returns "public" attributes of the DltException"""
        return {
            k: v
            for k, v in vars(self).items()
            if not k.startswith("_") and not callable(v) and not hasattr(self.__class__, k)
        }


class UnsupportedProcessStartMethodException(DltException):
    def __init__(self, method: str) -> None:
        self.method = method
        super().__init__(
            f"Process pool supports only fork start method, `{method}` not supported. Switch the"
            " pool type to threading"
        )


class CannotInstallDependencies(DltException):
    def __init__(self, dependencies: Sequence[str], interpreter: str, output: AnyStr) -> None:
        self.dependencies = dependencies
        self.interpreter = interpreter
        if isinstance(output, bytes):
            str_output = output.decode("utf-8")
        else:
            str_output = output
        super().__init__(
            f"Cannot install dependencies {', '.join(dependencies)} with `{interpreter=:}` and"
            f" pip:\n{str_output}\n"
        )


class VenvNotFound(DltException):
    def __init__(self, interpreter: str) -> None:
        self.interpreter = interpreter
        super().__init__(f"Venv with `{interpreter=:}` not found in path")


class ValueErrorWithKnownValues(ValueError):
    """Raise after receiving an invalid value and a set of valid values is known."""

    def __init__(self, key: str, value_received: Any, valid_values: Collection[Any]) -> None:
        self.key = key
        self.value_received = value_received
        self.valid_values = valid_values
        self.msg = (
            f"Received invalid value `{self.key}={self.value_received}`. "
            f"Valid values are: {self.valid_values}"
        )
        super().__init__(self.msg)


class TypeErrorWithKnownTypes(TypeError):
    """Raise after receiving an invalid value and a set of valid types is known."""

    def __init__(self, key: str, value_received: Any, valid_types: Collection[Any]) -> None:
        self.key = key
        self.value_received = value_received
        self.valid_types = valid_types
        self.msg = (
            f"Received invalid value `{self.key}={self.value_received}` of type"
            f" `{type(self.value_received).__name__}`. Valid types are: {self.valid_types}"
        )
        super().__init__(self.msg)


class TerminalException(BaseException):
    """
    Marks an exception that cannot be recovered from, should be mixed in into concrete exception class
    """


class TransientException(BaseException):
    """
    Marks an exception in operation that can be retried, should be mixed in into concrete exception class
    """


class TerminalValueError(ValueError, TerminalException):
    """
    ValueError that is unrecoverable
    """


class SignalReceivedException(KeyboardInterrupt, TerminalException):
    """Raises when signal comes. Derives from `BaseException` to not be caught in regular exception handlers."""

    def __init__(self, signal_code: int) -> None:
        self.signal_code = signal_code
        super().__init__(f"Received `{signal_code=:}`")


class DictValidationException(DltException):
    def __init__(
        self,
        msg: str,
        path: str,
        expected_type: Type[Any] = None,
        field: str = None,
        value: Any = None,
        nested_exceptions: List["DictValidationException"] = None,
    ) -> None:
        self.path = path
        self.expected_type = expected_type
        self.field = field
        self.value = value
        self.nested_exceptions = nested_exceptions
        self.msg = msg
        super().__init__(msg)

    def __str__(self) -> str:
        return f"Path `{self.path}`: " + self.msg


class ArgumentsOverloadException(DltException):
    def __init__(self, msg: str, func_name: str, *args: str) -> None:
        self.func_name = func_name
        msg = f"Arguments combination not allowed when calling function `{func_name}()`: {msg}"
        msg = "\n".join((msg, *args))
        super().__init__(msg)


class MissingDependencyException(DltException, ImportError):
    def __init__(self, caller: str, dependencies: Sequence[str], appendix: str = "") -> None:
        self.caller = caller
        self.dependencies = dependencies
        super().__init__(self._get_msg(appendix))

    def _get_msg(self, appendix: str) -> str:
        msg = f"""
You must install additional dependencies to run `{self.caller}`. If you use pip you may do the following:

{self._to_pip_install()}
"""
        if appendix:
            msg = msg + "\n" + appendix
        return msg

    def _to_pip_install(self) -> str:
        return "\n".join([f'pip install "{d}"' for d in self.dependencies])


class DependencyVersionException(DltException):
    def __init__(
        self, pkg_name: str, version_found: str, version_required: str, appendix: str = ""
    ) -> None:
        self.pkg_name = pkg_name
        self.version_found = version_found
        self.version_required = version_required
        super().__init__(self._get_msg(appendix))

    def _get_msg(self, appendix: str) -> str:
        msg = (
            f"Found `{self.pkg_name}=={self.version_found}`, while"
            f" `{self.pkg_name}{self.version_required}` is required."
        )
        if appendix:
            msg = msg + "\n" + appendix
        return msg


# NOTE MSSQL/Synapse is the only code using this
class SystemConfigurationException(DltException):
    pass


class PipelineException(DltException):
    def __init__(self, pipeline_name: str, msg: str) -> None:
        """Base class for all pipeline exceptions. Should not be raised."""
        self.pipeline_name = pipeline_name
        super().__init__(msg)


class PipelineStateNotAvailable(PipelineException):
    def __init__(self, source_state_key: Optional[str] = None) -> None:
        if source_state_key:
            msg = (
                f"The source `{source_state_key}` requested the access to pipeline state but no"
                " pipeline is active right now."
            )
        else:
            msg = (
                "The resource you called requested the access to pipeline state but no pipeline is"
                " active right now."
            )
        msg += (
            " Call `dlt.pipeline(...)` before you call the `@dlt.source` or  `@dlt.resource`"
            " decorated function."
        )
        self.source_state_key = source_state_key
        super().__init__(None, msg)


class ResourceNameNotAvailable(PipelineException):
    def __init__(self) -> None:
        super().__init__(
            None,
            "A resource state was requested but no active extract pipe context was found. Resource"
            " state may be only requested from `@dlt.resource` decorated function or with explicit"
            " resource name.",
        )


class SourceSectionNotAvailable(PipelineException):
    def __init__(self) -> None:
        msg = (
            "Access to state was requested without source section active. State should be requested"
            " from within the `@dlt.source` and `@dlt.resource` decorated function."
        )
        super().__init__(None, msg)



import inspect
import time

from abc import ABC, abstractmethod
from typing import (
    Any,
    Callable,
    ClassVar,
    Generic,
    Iterator,
    Mapping,
    Optional,
    Union,
    Dict,
    cast,
)

from dlt.common.data_writers.writers import count_rows_in_items
from dlt.common.schema.typing import TTableSchema
from dlt.common.typing import TAny, TDataItem, TDataItems, TypeVar

from dlt.extract.utils import (
    wrap_iterator,
)

from dlt.extract.items import SupportsPipe


ItemTransformFunctionWithMeta = Callable[[TDataItem, str], TAny]
ItemTransformFunctionNoMeta = Callable[[TDataItem], TAny]
ItemTransformFunc = Union[ItemTransformFunctionWithMeta[TAny], ItemTransformFunctionNoMeta[TAny]]

MetricsFunctionWithMeta = Callable[[TDataItems, Any, Dict[str, Any]], None]

TCustomMetrics = TypeVar(
    "TCustomMetrics", covariant=True, bound=Mapping[str, Any], default=Dict[str, Any]
)


class BaseItemTransform(Generic[TCustomMetrics]):
    def __init__(self) -> None:
        self._custom_metrics: TCustomMetrics = cast(TCustomMetrics, {})

    @property
    def custom_metrics(self) -> TCustomMetrics:
        return self._custom_metrics


class ItemTransform(BaseItemTransform[TCustomMetrics], ABC, Generic[TAny, TCustomMetrics]):
    _f_meta: ItemTransformFunctionWithMeta[TAny] = None
    _f: ItemTransformFunctionNoMeta[TAny] = None

    placement_affinity: ClassVar[float] = 0
    """Tell how strongly an item sticks to start (-1) or end (+1) of pipe."""

    def __init__(self, transform_f: ItemTransformFunc[TAny]) -> None:
        super().__init__()
        # inspect the signature
        sig = inspect.signature(transform_f)
        # TODO: use TypeGuard here to get rid of type ignore
        if len(sig.parameters) == 1:
            self._f = transform_f  # type: ignore
        else:  # TODO: do better check
            self._f_meta = transform_f  # type: ignore

    def bind(
        self: "ItemTransform[TAny, TCustomMetrics]", pipe: SupportsPipe
    ) -> "ItemTransform[TAny, TCustomMetrics]":
        return self

    @abstractmethod
    def __call__(self, item: TDataItems, meta: Any = None) -> Optional[TDataItems]:
        """Transforms `item` (a list of TDataItem or a single TDataItem) and returns or yields TDataItems. Returns None to consume item (filter out)"""
        pass


class FilterItem(ItemTransform[bool, Dict[str, Any]]):
    # mypy needs those to type correctly
    _f_meta: ItemTransformFunctionWithMeta[bool]
    _f: ItemTransformFunctionNoMeta[bool]

    def __call__(self, item: TDataItems, meta: Any = None) -> Optional[TDataItems]:
        if isinstance(item, list):
            # preserve type of empty lists
            if len(item) == 0:
                return item

            if self._f_meta:
                item = [i for i in item if self._f_meta(i, meta)]
            else:
                item = [i for i in item if self._f(i)]
            if not item:
                # item was fully consumed by the filter
                return None
            return item
        else:
            if self._f_meta:
                return item if self._f_meta(item, meta) else None
            else:
                return item if self._f(item) else None


class MapItem(ItemTransform[TDataItem, Dict[str, Any]]):
    # mypy needs those to type correctly
    _f_meta: ItemTransformFunctionWithMeta[TDataItem]
    _f: ItemTransformFunctionNoMeta[TDataItem]

    def __call__(self, item: TDataItems, meta: Any = None) -> Optional[TDataItems]:
        if isinstance(item, list):
            # preserve type of empty lists
            if len(item) == 0:
                return item

            if self._f_meta:
                return [self._f_meta(i, meta) for i in item]
            else:
                return [self._f(i) for i in item]
        else:
            if self._f_meta:
                return self._f_meta(item, meta)
            else:
                return self._f(item)


class YieldMapItem(ItemTransform[Iterator[TDataItem], Dict[str, Any]]):
    # mypy needs those to type correctly
    _f_meta: ItemTransformFunctionWithMeta[TDataItem]
    _f: ItemTransformFunctionNoMeta[TDataItem]

    def __call__(self, item: TDataItems, meta: Any = None) -> Optional[TDataItems]:
        if isinstance(item, list):
            # preserve type of empty lists
            if len(item) == 0:
                yield item

            for i in item:
                if self._f_meta:
                    yield from self._f_meta(i, meta)
                else:
                    yield from self._f(i)
        else:
            if self._f_meta:
                yield from self._f_meta(item, meta)
            else:
                yield from self._f(item)


class ValidateItem(ItemTransform[TDataItem, Dict[str, Any]]):
    """Base class for validators of data items.

    Subclass should implement the `__call__` method to either return the data item(s) or raise `extract.exceptions.ValidationError`.
    See `PydanticValidator` for possible implementation.
    """

    placement_affinity: ClassVar[float] = 0.9  # stick to end but less than incremental

    table_name: str

    def bind(self, pipe: SupportsPipe) -> ItemTransform[TDataItem, Dict[str, Any]]:
        self.table_name = pipe.name
        return self

    def compute_table_schema(
        self, item: TDataItem = None, meta: Any = None
    ) -> Optional[TTableSchema]:
        """Computes table schema enforced by this validator. For dynamic (item dependent) or
        schemas with discriminator field use item and meta.

        Return None if no authoratitative schema can be computed
        """
        return None


class LimitItem(ItemTransform[TDataItem, Dict[str, Any]]):
    placement_affinity: ClassVar[float] = 1.1  # stick to end right behind incremental

    def __init__(
        self, max_items: Optional[int], max_time: Optional[float], count_rows: bool
    ) -> None:
        BaseItemTransform.__init__(self)
        self.max_items = max_items if max_items is not None else -1
        self.max_time = max_time
        self.count_rows = count_rows

    def bind(self, pipe: SupportsPipe) -> "LimitItem":
        # we also wrap iterators to make them stoppable
        if isinstance(pipe.gen, Iterator):
            pipe.replace_gen(wrap_iterator(pipe.gen))

        self.gen = pipe.gen
        self.count = 0
        self.exhausted = False
        self.start_time = time.monotonic()

        return self

    def limit(self, chunk_size: int) -> Optional[int]:
        """Calculate the maximum number of rows to which result is limited. Limit works in chunks
        that controlled by the data source and this must be provided in `chunk_size`.
        `chunk_size` will be ignore if counting rows (`count_rows` is `True`). Mind that
        this count method will not split batches so you may get more items (up to the full last batch)
        than `limit` method indicates.
        """
        if self.max_items in (None, -1):
            return None
        return self.max_items * (1 if self.count_rows else chunk_size)

    def __call__(self, item: TDataItems, meta: Any = None) -> Optional[TDataItems]:
        # do not count None
        if item is None:
            return None

        # do not return any late arriving items
        if self.exhausted:
            return None

        if self.count_rows:
            self.count += count_rows_in_items(item)
        else:
            # NOTE: we count empty batches/pages in this mode
            self.count += 1

        # detect when the limit is reached, max time or yield count
        if (
            (self.count >= self.max_items and self.max_items >= 0)
            or (self.max_time and time.monotonic() - self.start_time > self.max_time)
            or self.max_items == 0
        ):
            self.exhausted = True
            if inspect.isgenerator(self.gen):
                self.gen.close()

            # if max items is not 0, we return the last item
            # otherwise never return anything
            if self.max_items != 0:
                return item
            return None

        return item


class MetricsItem(ItemTransform[None, Dict[str, Any]]):
    """Collects custom metrics from data flowing through the pipe without modifying items.

    The metrics function receives items, optionally meta, and a metrics dictionary that it can
    update in-place. The items are passed through unchanged.
    """

    _metrics_f: MetricsFunctionWithMeta = None

    def __init__(self, metrics_f: MetricsFunctionWithMeta) -> None:
        BaseItemTransform.__init__(self)
        self._metrics_f = metrics_f

    def __call__(self, item: TDataItems, meta: Any = None) -> Optional[TDataItems]:
        self._metrics_f(item, meta, self._custom_metrics)
        return item



from typing import Tuple, Dict, Optional

from dlt.common import logger
from dlt.common.destination import PreparedTableSchema
from dlt.common.schema.typing import (
    TColumnSchema,
    TDataType,
    TColumnType,
)
from dlt.common.destination.capabilities import DataTypeMapper
from dlt.common.typing import TLoaderFileFormat
from dlt.common.utils import without_none


# NOTE it seems that multiple methods pass `table` only for logging/exception purposes
# should refactor to remove `table` kwarg and have callers catch exceptions
class TypeMapperImpl(DataTypeMapper):
    sct_to_unbound_dbt: Dict[TDataType, str]
    """Data types without precision or scale specified (e.g. `"text": "varchar"` in postgres)"""
    sct_to_dbt: Dict[TDataType, str]
    """Data types that require a precision or scale (e.g. `"text": "varchar(%i)"` or `"decimal": "numeric(%i,%i)"` in postgres).
    Values should have printf placeholders for precision (and scale if applicable)
    """

    dbt_to_sct: Dict[str, TDataType]

    def ensure_supported_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema,
        loader_file_format: TLoaderFileFormat,
    ) -> None:
        pass

    def to_db_integer_type(self, column: TColumnSchema, table: PreparedTableSchema = None) -> str:
        # Override in subclass if db supports other integer types (e.g. smallint, integer, tinyint, etc.)
        return self.sct_to_unbound_dbt["bigint"]

    def to_db_datetime_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema = None,
    ) -> str:
        # Override in subclass if db supports other timestamp types (e.g. with different time resolutions)
        timezone = column.get("timezone")
        precision = column.get("precision")

        if timezone is not None or precision is not None:
            message = (
                "Column flags for timezone or precision are not yet supported in this"
                " destination. One or both of these flags were used in column"
                f" '{column.get('name')}'."
            )
            # TODO: refactor lancedb and weaviate to make table object required
            if table:
                message += f" in table '{table.get('name')}'."

            logger.warning(message)

        return None

    def to_db_time_type(self, column: TColumnSchema, table: PreparedTableSchema = None) -> str:
        # Override in subclass if db supports other time types (e.g. with different time resolutions)
        return None

    def to_db_decimal_type(self, column: TColumnSchema) -> str:
        precision_tup = self.decimal_precision(column.get("precision"), column.get("scale"))
        if not precision_tup or "decimal" not in self.sct_to_dbt:
            return self.sct_to_unbound_dbt["decimal"]
        return self.sct_to_dbt["decimal"] % (precision_tup[0], precision_tup[1])

    # TODO: refactor lancedb and weaviate to make table object required
    def to_destination_type(self, column: TColumnSchema, table: PreparedTableSchema) -> str:
        sc_t = column["data_type"]
        if sc_t == "bigint":
            db_t = self.to_db_integer_type(column, table)
        elif sc_t == "timestamp":
            db_t = self.to_db_datetime_type(column, table)
        elif sc_t == "time":
            db_t = self.to_db_time_type(column, table)
        elif sc_t == "decimal":
            db_t = self.to_db_decimal_type(column)
        else:
            db_t = None
        if db_t:
            return db_t
        # try templates with precision
        bounded_template = self.sct_to_dbt.get(sc_t)
        if not bounded_template:
            return self.sct_to_unbound_dbt[sc_t]
        precision_tuple = self.precision_tuple_or_default(sc_t, column)
        if not precision_tuple:
            return self.sct_to_unbound_dbt[sc_t]
        return self.sct_to_dbt[sc_t] % precision_tuple

    def precision_tuple_or_default(
        self, data_type: TDataType, column: TColumnSchema
    ) -> Optional[Tuple[int, ...]]:
        precision = column.get("precision")
        scale = column.get("scale")
        if data_type in ("timestamp", "time"):
            if precision is None:
                return None  # Use default which is usually the max
        elif data_type == "decimal":
            return self.decimal_precision(precision, scale)
        elif data_type == "wei":
            return self.wei_precision(precision, scale)

        if precision is None:
            return None
        elif scale is None:
            return (precision,)
        return (precision, scale)

    def decimal_precision(
        self, precision: Optional[int] = None, scale: Optional[int] = None
    ) -> Optional[Tuple[int, int]]:
        defaults = self.capabilities.decimal_precision
        if not defaults:
            return None
        default_precision, default_scale = defaults
        return (
            precision if precision is not None else default_precision,
            scale if scale is not None else default_scale,
        )

    def wei_precision(
        self, precision: Optional[int] = None, scale: Optional[int] = None
    ) -> Optional[Tuple[int, int]]:
        defaults = self.capabilities.wei_precision
        if not defaults:
            return None
        default_precision, default_scale = defaults
        return (
            precision if precision is not None else default_precision,
            scale if scale is not None else default_scale,
        )

    def from_destination_type(
        self, db_type: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        return without_none(
            dict(  # type: ignore[return-value]
                data_type=self.dbt_to_sct.get(db_type, "text"), precision=precision, scale=scale
            )
        )



"""Data quality dashboard helpers: controls, section widget, and raw table display."""

import traceback
from typing import Callable, List, Optional, Tuple

import marimo as mo

import dlt
import pyarrow

from dlt._workspace.helpers.dashboard import strings
from dlt._workspace.helpers.dashboard.utils import queries, ui


def create_dq_controls(
    pipeline: dlt.Pipeline,
) -> Tuple[
    Optional[mo.ui.checkbox],
    Optional[mo.ui.dropdown],
    Optional[mo.ui.slider],
    Optional[pyarrow.Table],
]:
    """Create data quality filter controls by importing from dlthub.

    Returns (show_failed_filter, table_filter, rate_filter, checks_arrow).
    All None on failure or if dlthub is not available.
    """
    try:
        from dlthub.data_quality._dashboard import create_data_quality_controls

        result: Tuple[
            Optional[mo.ui.checkbox],
            Optional[mo.ui.dropdown],
            Optional[mo.ui.slider],
            Optional[pyarrow.Table],
        ] = create_data_quality_controls(pipeline)
        return result
    except Exception:
        return None, None, None, None


def build_dq_section(
    pipeline: dlt.Pipeline,
    show_failed_filter: Optional[mo.ui.checkbox],
    table_filter: Optional[mo.ui.dropdown],
    rate_filter: Optional[mo.ui.slider],
    checks_arrow: Optional[pyarrow.Table],
) -> Tuple[List[mo.Html], Optional[mo.ui.switch]]:
    """Build the data quality widget section.

    Returns (result_widgets, raw_table_switch). raw_table_switch is None when
    there is no data or on error.
    """
    result: List[mo.Html] = []
    raw_table_switch: Optional[mo.ui.switch] = None

    try:
        from dlthub.data_quality._dashboard import data_quality_widget

        # extract values from controls
        show_failed_value = show_failed_filter.value if show_failed_filter is not None else False
        table_value = None
        if table_filter is not None and table_filter.value != "All":
            table_value = table_filter.value
        rate_value = rate_filter.value if rate_filter is not None else None

        widget_output = data_quality_widget(
            dlt_pipeline=pipeline,
            failure_rate_slider=rate_filter,
            failure_rate_filter_value=rate_value,
            show_only_failed_checkbox=show_failed_filter,
            show_only_failed_value=show_failed_value,
            table_dropdown=table_filter,
            table_name_filter_value=table_value,
            checks_arrow=checks_arrow,
        )
        if widget_output is not None:
            result.append(widget_output)

        # only show raw table switch if there is data
        if checks_arrow is not None and checks_arrow.num_rows > 0:
            raw_table_switch = mo.ui.switch(
                value=False,
                label=ui.small(strings.data_quality_show_raw_table),
            )
            result.append(mo.hstack([raw_table_switch], justify="start"))
    except ImportError:
        result.append(mo.md(strings.data_quality_not_available))
    except Exception as exc:
        result.append(
            ui.error_callout(
                strings.data_quality_error_loading.format(exc),
            )
        )

    return result, raw_table_switch


def build_dq_raw_table(
    pipeline: dlt.Pipeline,
    get_result: Callable[[], pyarrow.Table],
    set_result: Callable[[pyarrow.Table], None],
) -> List[mo.Html]:
    """Build the raw data quality checks table.

    Returns list of widgets to display.
    """
    result: List[mo.Html] = []

    try:
        from dlthub import data_quality as dq

        _error_message: str = None
        with mo.status.spinner(title=strings.data_quality_loading_raw_table_spinner):
            try:
                _raw_sql_query = dq.read_check(pipeline.dataset())
                _raw_query_result, _error_message, _traceback_string = queries.get_query_result(
                    pipeline, _raw_sql_query.to_sql()
                )
                set_result(_raw_query_result)
            except Exception as exc:
                _error_message = str(exc)
                _traceback_string = traceback.format_exc()

        if _error_message:
            result.append(
                ui.error_callout(
                    strings.data_quality_raw_table_error.format(_error_message),
                    traceback_string=_traceback_string,
                )
            )

        if (_last_result := get_result()) is not None:
            result.append(ui.dlt_table(_last_result, freeze_column=None))
    except ImportError:
        result.append(
            mo.callout(
                mo.md(strings.data_quality_raw_table_not_available),
                kind="warn",
            )
        )
    except Exception as exc:
        result.append(
            ui.error_callout(
                strings.data_quality_raw_table_error.format(exc),
            )
        )

    return result



"""dltHub telemetry using using anonymous tracker"""

# several code fragments come from https://github.com/RasaHQ/rasa/blob/main/rasa/telemetry.py
import contextlib
import os
import base64
from typing import Iterator, Literal, Optional
from requests import Session

from dlt.common import logger
from dlt.common.managed_thread_pool import ManagedThreadPool
from dlt.common.configuration.specs import RuntimeConfiguration
from dlt.common.runtime.exec_info import get_execution_context, TExecutionContext, run_context_name
from dlt.common.runtime import run_context
from dlt.common.typing import DictStrAny, StrAny
from dlt.common.utils import uniq_id

from dlt.version import __version__

TEventCategory = Literal["pipeline", "command", "helper", "mcp"]

_THREAD_POOL: ManagedThreadPool = None
_WRITE_KEY: str = None
_REQUEST_TIMEOUT = (1.0, 1.0)  # short connect & send timeouts
_ANON_TRACKER_ENDPOINT: str = None
_TRACKER_CONTEXT: TExecutionContext = None
requests: Session = None


def init_anon_tracker(config: RuntimeConfiguration) -> None:
    if config.dlthub_telemetry_endpoint is None:
        raise ValueError("`dlthub_telemetry_endpoint` not specified in `RunConfiguration`")

    if config.dlthub_telemetry_endpoint == "https://api.segment.io/v1/track":
        assert (
            config.dlthub_telemetry_segment_write_key
        ), "dlthub_telemetry_segment_write_key not present in RunConfiguration"

    # lazily import requests to avoid binding config before initialization
    global requests
    from dlt.sources.helpers.requests import Client

    # fail fast, don't block user
    requests = Client(request_timeout=_REQUEST_TIMEOUT, request_max_attempts=0)  # type: ignore[assignment]

    global _WRITE_KEY, _ANON_TRACKER_ENDPOINT, _THREAD_POOL
    # start the pool
    # create thread pool to send telemetry to anonymous tracker
    if _THREAD_POOL is None:
        _THREAD_POOL = ManagedThreadPool("anon_tracker", 1)
        # do not instantiate lazy: this happens in _future_send which may be triggered
        # from a thread and also in parallel when many pipelines are run at once
        _THREAD_POOL._create_thread_pool()
    # store write key if present
    if config.dlthub_telemetry_segment_write_key:
        key_bytes = (config.dlthub_telemetry_segment_write_key + ":").encode("ascii")
        _WRITE_KEY = base64.b64encode(key_bytes).decode("utf-8")
    # store endpoint
    _ANON_TRACKER_ENDPOINT = config.dlthub_telemetry_endpoint
    # cache the tracker context
    _default_context_fields()


def disable_anon_tracker() -> None:
    global _WRITE_KEY, _TRACKER_CONTEXT, _ANON_TRACKER_ENDPOINT, _THREAD_POOL
    if _THREAD_POOL is not None:
        _THREAD_POOL.stop(True)
    _ANON_TRACKER_ENDPOINT = None
    _WRITE_KEY = None
    _TRACKER_CONTEXT = None
    _THREAD_POOL = None


def track(event_category: TEventCategory, event_name: str, properties: DictStrAny) -> None:
    """Tracks a telemetry event.

    The tracker event name will be created as "{event_category}_{event_name}

    Args:
        event_category: Category of the event: pipeline or cli
        event_name: Name of the event.
        properties: Dictionary containing the event's properties.
    """
    if properties is None:
        properties = {}

    properties.update({"event_category": event_category, "event_name": event_name})

    try:
        context = _default_context_fields()
        # always refresh run context name, it may change at runtime
        context["run_context"] = run_context_name()
        _send_event(f"{event_category}_{event_name}", properties, context)
    except Exception as e:
        logger.debug(f"Skipping telemetry reporting: {e}")
        raise


def before_send(event: DictStrAny) -> Optional[DictStrAny]:
    """Called before sending event. Does nothing, patch this function in the module for custom behavior"""
    return event


@contextlib.contextmanager
def always_track() -> Iterator[None]:
    # if anon tracker was disabled
    disable_after = _ANON_TRACKER_ENDPOINT is None
    if disable_after:
        from dlt.common.configuration.container import Container
        from dlt.common.configuration.specs.pluggable_run_context import PluggableRunContext

        init_anon_tracker(Container()[PluggableRunContext].context.runtime_config)
    try:
        yield
    finally:
        if disable_after:
            disable_anon_tracker()


def _tracker_request_header(write_key: str) -> StrAny:
    """Use a segment write key to create authentication headers for the segment API.

    Args:
        write_key: Authentication key for segment.

    Returns:
        Authentication headers for segment.
    """
    headers = {"Content-Type": "application/json"}
    if write_key:
        headers["Authorization"] = "Basic {}".format(write_key)
    return headers


def get_anonymous_id() -> str:
    """Creates or reads a anonymous user id"""
    home_dir = run_context.active().global_dir

    if not os.path.isdir(home_dir):
        os.makedirs(home_dir, exist_ok=True)
    anonymous_id_file = os.path.join(home_dir, ".anonymous_id")
    if not os.path.isfile(anonymous_id_file):
        anonymous_id = uniq_id()
        with open(anonymous_id_file, "w", encoding="utf-8") as f:
            f.write(anonymous_id)
    else:
        with open(anonymous_id_file, "r", encoding="utf-8") as f:
            anonymous_id = f.read()
    return anonymous_id


def _create_request_payload(event_name: str, properties: StrAny, context: StrAny) -> DictStrAny:
    """Compose a valid payload for the tracker.

    Args:
        event_name: Name of the event.
        properties: Values to report along the event.
        context: Context information about the event.

    Returns:
        Valid tracker payload.
    """
    return {
        "anonymousId": get_anonymous_id(),
        "event": event_name,
        "properties": properties,
        "context": context,
    }


def _default_context_fields() -> TExecutionContext:
    """Return a dictionary that contains the default context values.

    Return:
        A new context containing information about the runtime environment.
    """
    global _TRACKER_CONTEXT

    if not _TRACKER_CONTEXT:
        # Make sure to update the example in docs/reference/telemetry.md
        # if you change / add context
        _TRACKER_CONTEXT = get_execution_context()

    # avoid returning the cached dict --> caller could modify the dictionary...
    # usually we would use `lru_cache`, but that doesn't return a dict copy and
    # doesn't work on inner functions, so we need to roll our own caching...
    return _TRACKER_CONTEXT.copy()


def _send_event(event_name: str, properties: StrAny, context: StrAny) -> None:
    """Report the contents of an event to the tracker endpoint.

    Args:
        event_name: Name of the event.
        properties: Values to report along the event.
        context: Context information about the event.
    """
    # formulate payload and process in before send
    payload = before_send(_create_request_payload(event_name, properties, context))
    # skip empty payloads
    if not payload:
        logger.debug("Skipping request to external service: payload was filtered out.")
        return

    if _ANON_TRACKER_ENDPOINT is None:
        logger.debug("Skipping request to external service: telemetry endpoint not set.")
        return

    headers = _tracker_request_header(_WRITE_KEY)

    def _future_send() -> None:
        import time

        start_ts = time.time_ns()
        resp = requests.post(
            _ANON_TRACKER_ENDPOINT, headers=headers, json=payload, timeout=_REQUEST_TIMEOUT
        )
        end_ts = time.time_ns()
        elapsed_time = (end_ts - start_ts) / 10e6
        # print(f"SENDING TO TRACKER done: {elapsed_time}ms Status: {resp.status_code}")
        # handle different failure cases
        if resp.status_code not in [200, 204]:
            logger.debug(
                f"Tracker request returned a {resp.status_code} response in {elapsed_time}ms. Body:"
                f" {resp.text}"
            )
        else:
            logger.debug(
                f"Tracker request returned a {resp.status_code} response in {elapsed_time}ms.s"
            )
            if resp.status_code == 200:
                # parse the response if available
                data = resp.json()
                if not data.get("success"):
                    logger.debug(f"Tracker telemetry request returned a failure. Response: {data}")

    _THREAD_POOL.thread_pool.submit(_future_send)



from dlt.common.exceptions import DltException


class RuntimeException(DltException):
    pass


class RunContextNotAvailable(RuntimeException):
    def __init__(self, run_dir: str, msg: str):
        super().__init__(run_dir, msg)



from types import TracebackType
from typing import Optional, Sequence, List, Dict, Type, Iterable, Any
import threading

from dlt.common import logger
from dlt.common.json import json
from dlt.common.pendulum import pendulum
from dlt.common.schema import Schema, TSchemaTables
from dlt.common.schema.typing import C_DLT_LOAD_ID, C_DLT_LOADS_TABLE_LOAD_ID
from dlt.common.schema.utils import (
    get_columns_names_with_prop,
    loads_table,
    normalize_table_identifiers,
    version_table,
)
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.destination.client import (
    PreparedTableSchema,
    RunnableLoadJob,
    JobClientBase,
    WithStateSync,
    LoadJob,
)
from dlt.common.destination.exceptions import DestinationUndefinedEntity

from dlt.common.storages import FileStorage

from dlt.destinations.job_client_impl import StorageSchemaInfo, StateInfo
from dlt.destinations.utils import get_pipeline_state_query_columns
from dlt.destinations.impl.qdrant.configuration import QdrantClientConfiguration
from dlt.destinations.impl.qdrant.qdrant_adapter import VECTORIZE_HINT

from qdrant_client import QdrantClient as QC, models
from qdrant_client.qdrant_fastembed import uuid
from qdrant_client.http.exceptions import UnexpectedResponse


class QDrantLoadJob(RunnableLoadJob):
    def __init__(
        self,
        file_path: str,
        client_config: QdrantClientConfiguration,
        collection_name: str,
    ) -> None:
        super().__init__(file_path)
        self._collection_name = collection_name
        self._config = client_config
        self._job_client: "QdrantClient" = None

    def run(self) -> None:
        embedding_fields = get_columns_names_with_prop(self._load_table, VECTORIZE_HINT)
        unique_identifiers = self._list_unique_identifiers(self._load_table)
        with FileStorage.open_zipsafe_ro(self._file_path) as f:
            ids: List[str]
            docs, payloads, ids = [], [], []

            for line in f:
                data = json.loads(line)
                point_id = (
                    self._generate_uuid(data, unique_identifiers, self._collection_name)
                    if unique_identifiers
                    else str(uuid.uuid4())
                )
                payloads.append(data)
                ids.append(point_id)
                if len(embedding_fields) > 0:
                    docs.append(self._get_embedding_doc(data, embedding_fields))

            if len(embedding_fields) > 0:
                embedding_model = self._job_client.db_client._get_or_init_model(
                    self._job_client.db_client.embedding_model_name
                )
                embeddings: List[Any] = list(
                    embedding_model.embed(
                        docs,
                        batch_size=self._config.embedding_batch_size,
                        parallel=self._config.embedding_parallelism,
                    )
                )
                vector_name = self._job_client.db_client.get_vector_field_name()
                embeddings = [{vector_name: embedding.tolist()} for embedding in embeddings]
            else:
                embeddings = [{}] * len(ids)
            assert len(embeddings) == len(payloads) == len(ids)

            self._upload_data(vectors=embeddings, ids=ids, payloads=payloads)

    def _get_embedding_doc(self, data: Dict[str, Any], embedding_fields: List[str]) -> str:
        """Returns a document to generate embeddings for.

        Args:
            data (Dict[str, Any]): A dictionary of data to be loaded.

        Returns:
            str: A concatenated string of all the fields intended for embedding.
        """
        doc = "\n".join(str(data[key]) for key in embedding_fields)
        return doc

    def _list_unique_identifiers(self, table_schema: PreparedTableSchema) -> Sequence[str]:
        """Returns a list of unique identifiers for a table.

        Args:
            table_schema (PreparedTableSchema): a dlt table schema.

        Returns:
            Sequence[str]: A list of unique column identifiers.
        """
        if table_schema.get("write_disposition") == "merge":
            primary_keys = get_columns_names_with_prop(table_schema, "primary_key")
            if primary_keys:
                return primary_keys
        return get_columns_names_with_prop(table_schema, "unique")

    def _upload_data(
        self, ids: Iterable[Any], vectors: Iterable[Any], payloads: Iterable[Any]
    ) -> None:
        """Uploads data to a Qdrant instance in a batch. Supports retries and parallelism.

        Args:
            ids (Iterable[Any]): Point IDs to be uploaded to the collection
            vectors (Iterable[Any]): Embeddings to be uploaded to the collection
            payloads (Iterable[Any]): Payloads to be uploaded to the collection
        """
        self._job_client.db_client.upload_collection(
            self._collection_name,
            ids=ids,
            payload=payloads,
            vectors=vectors,
            parallel=self._config.upload_parallelism,
            batch_size=self._config.upload_batch_size,
            max_retries=self._config.upload_max_retries,
        )

    def _generate_uuid(
        self, data: Dict[str, Any], unique_identifiers: Sequence[str], collection_name: str
    ) -> str:
        """Generates deterministic UUID. Used for deduplication.

        Args:
            data (Dict[str, Any]): Arbitrary data to generate UUID for.
            unique_identifiers (Sequence[str]): A list of unique identifiers.
            collection_name (str): Qdrant collection name.

        Returns:
            str: A string representation of the generated UUID
        """
        data_id = "_".join(str(data[key]) for key in unique_identifiers)
        return str(uuid.uuid5(uuid.NAMESPACE_DNS, collection_name + data_id))


class QdrantClient(JobClientBase, WithStateSync):
    """Qdrant Destination Handler"""

    def __init__(
        self,
        schema: Schema,
        config: QdrantClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        super().__init__(schema, config, capabilities)
        # get definitions of the dlt tables, normalize column names and keep for later use
        version_table_ = normalize_table_identifiers(version_table(), schema.naming)
        self.version_collection_properties = list(version_table_["columns"].keys())
        loads_table_ = normalize_table_identifiers(loads_table(), schema.naming)
        self.loads_collection_properties = list(loads_table_["columns"].keys())
        state_table_ = normalize_table_identifiers(
            get_pipeline_state_query_columns(), schema.naming
        )
        self.pipeline_state_properties = list(state_table_["columns"].keys())

        self.config: QdrantClientConfiguration = config
        self.db_client: QC = None
        self.model = config.model

    @property
    def dataset_name(self) -> str:
        return self.config.normalize_dataset_name(self.schema)

    @property
    def sentinel_collection(self) -> str:
        return self.dataset_name or "DltSentinelCollection"

    def _make_qualified_collection_name(self, table_name: str) -> str:
        """Generates a qualified collection name.

        Args:
            table_name (str): Name of the table.

        Returns:
            str: The dataset name and table name concatenated with a separator if dataset name is present.
        """
        dataset_separator = self.config.dataset_separator
        return (
            f"{self.dataset_name}{dataset_separator}{table_name}"
            if self.dataset_name
            else table_name
        )

    def _create_collection(self, full_collection_name: str) -> None:
        """Creates a collection in Qdrant.

        Args:
            full_collection_name (str): The name of the collection to be created.
        """

        # Generates config for a named vector according to the selected model.
        # Eg: vector_config={
        #     "fast-bge-small-en": {
        #       "size": 364,
        #       "distance": "Cosine"
        #     },
        # }
        vectors_config = self.db_client.get_fastembed_vector_params()

        self.db_client.create_collection(
            collection_name=full_collection_name, vectors_config=vectors_config
        )
        # TODO: we can use index hints to create indexes on properties or full text
        # self.db_client.create_payload_index(full_collection_name, "_dlt_load_id", field_type="float")

    def _create_point_no_vector(self, obj: Dict[str, Any], collection_name: str) -> None:
        """Inserts a point into a Qdrant collection without a vector.

        Args:
            obj (Dict[str, Any]): The arbitrary data to be inserted as payload.
            collection_name (str): The name of the collection to insert the point into.
        """
        self.db_client.upsert(
            collection_name,
            points=[
                models.PointStruct(
                    id=str(uuid.uuid4()),
                    payload=obj,
                    vector={},
                )
            ],
        )

    def drop_storage(self) -> None:
        """Drop the dataset from the Qdrant instance.

        Deletes all collections in the dataset and all data associated.
        Deletes the sentinel collection.

        If dataset name was not provided, it deletes all the tables in the current schema
        """
        collections = self.db_client.get_collections().collections
        collection_name_list = [collection.name for collection in collections]

        if self.dataset_name:
            prefix = f"{self.dataset_name}{self.config.dataset_separator}"

            for collection_name in collection_name_list:
                if collection_name.startswith(prefix):
                    self.db_client.delete_collection(collection_name)
        else:
            for collection_name in self.schema.tables.keys():
                if collection_name in collection_name_list:
                    self.db_client.delete_collection(collection_name)

        self._delete_sentinel_collection()

    def initialize_storage(self, truncate_tables: Iterable[str] = None) -> None:
        if not self.is_storage_initialized():
            self._create_sentinel_collection()
        elif truncate_tables:
            for table_name in truncate_tables:
                qualified_table_name = self._make_qualified_collection_name(table_name=table_name)
                if self._collection_exists(qualified_table_name):
                    continue

                self.db_client.delete_collection(qualified_table_name)
                self._create_collection(full_collection_name=qualified_table_name)

    def is_storage_initialized(self) -> bool:
        return self._collection_exists(self.sentinel_collection, qualify_table_name=False)

    def _create_sentinel_collection(self) -> None:
        """Create an empty collection to indicate that the storage is initialized."""
        self._create_collection(full_collection_name=self.sentinel_collection)

    def _delete_sentinel_collection(self) -> None:
        """Delete the sentinel collection."""
        self.db_client.delete_collection(self.sentinel_collection)

    def update_stored_schema(
        self,
        only_tables: Iterable[str] = None,
        expected_update: TSchemaTables = None,
    ) -> Optional[TSchemaTables]:
        applied_update = super().update_stored_schema(only_tables, expected_update)
        schema_info = self.get_stored_schema_by_hash(self.schema.stored_version_hash)
        if schema_info is None:
            logger.info(
                f"Schema with hash {self.schema.stored_version_hash} "
                "not found in the storage. upgrading"
            )
            self._execute_schema_update(only_tables)
        else:
            logger.info(
                f"Schema with hash {self.schema.stored_version_hash} "
                f"inserted at {schema_info.inserted_at} found "
                "in storage, no upgrade required"
            )
        # we assume that expected_update == applied_update so table schemas in dest were not
        # externally changed
        return applied_update

    def get_stored_state(self, pipeline_name: str) -> Optional[StateInfo]:
        """Loads compressed state from destination storage
        By finding a load id that was completed
        """
        # normalize property names
        p_load_id = self.schema.naming.normalize_identifier(C_DLT_LOADS_TABLE_LOAD_ID)
        p_dlt_load_id = self.schema.naming.normalize_identifier(C_DLT_LOAD_ID)
        p_pipeline_name = self.schema.naming.normalize_identifier("pipeline_name")
        p_created_at = self.schema.naming.normalize_identifier("created_at")

        limit = 10
        start_from = None
        while True:
            try:
                scroll_table_name = self._make_qualified_collection_name(
                    self.schema.state_table_name
                )
                state_records, offset = self.db_client.scroll(
                    scroll_table_name,
                    with_payload=self.pipeline_state_properties,
                    scroll_filter=models.Filter(
                        must=[
                            models.FieldCondition(
                                key=p_pipeline_name, match=models.MatchValue(value=pipeline_name)
                            )
                        ]
                    ),
                    order_by=models.OrderBy(
                        key=p_created_at,
                        direction=models.Direction.DESC,
                        start_from=start_from,
                    ),
                    limit=limit,
                )
                if len(state_records) == 0:
                    return None
                for state_record in state_records:
                    state = state_record.payload
                    start_from = state[p_created_at]
                    load_id = state[p_dlt_load_id]
                    scroll_table_name = self._make_qualified_collection_name(
                        self.schema.loads_table_name
                    )
                    load_records = self.db_client.count(
                        scroll_table_name,
                        exact=True,
                        count_filter=models.Filter(
                            must=[
                                models.FieldCondition(
                                    key=p_load_id, match=models.MatchValue(value=load_id)
                                )
                            ]
                        ),
                    )
                    if load_records.count == 0:
                        continue
                    return StateInfo.from_normalized_mapping(state, self.schema.naming)
            except UnexpectedResponse as e:
                if e.status_code == 404:
                    raise DestinationUndefinedEntity(str(e)) from e
                raise
            except ValueError as e:  # Local qdrant error
                if "not found" in str(e):
                    raise DestinationUndefinedEntity(str(e)) from e
                raise

    def get_stored_schema(self, schema_name: str = None) -> Optional[StorageSchemaInfo]:
        """Retrieves newest schema from destination storage"""
        try:
            scroll_table_name = self._make_qualified_collection_name(self.schema.version_table_name)
            p_schema_name = self.schema.naming.normalize_identifier("schema_name")
            p_inserted_at = self.schema.naming.normalize_identifier("inserted_at")

            name_filter = (
                models.Filter(
                    must=[
                        models.FieldCondition(
                            key=p_schema_name,
                            match=models.MatchValue(value=schema_name),
                        )
                    ]
                )
                if schema_name
                else None
            )

            response = self.db_client.scroll(
                scroll_table_name,
                with_payload=True,
                scroll_filter=name_filter,
                limit=1,
                order_by=models.OrderBy(
                    key=p_inserted_at,
                    direction=models.Direction.DESC,
                ),
            )
            if not response[0]:
                return None
            payload = response[0][0].payload
            return StorageSchemaInfo.from_normalized_mapping(payload, self.schema.naming)
        except UnexpectedResponse as e:
            if e.status_code == 404:
                raise DestinationUndefinedEntity(str(e)) from e
            raise
        except ValueError as e:  # Local qdrant error
            if "not found" in str(e):
                raise DestinationUndefinedEntity(str(e)) from e
            raise

    def get_stored_schema_by_hash(self, schema_hash: str) -> Optional[StorageSchemaInfo]:
        try:
            scroll_table_name = self._make_qualified_collection_name(self.schema.version_table_name)
            p_version_hash = self.schema.naming.normalize_identifier("version_hash")
            response = self.db_client.scroll(
                scroll_table_name,
                with_payload=True,
                scroll_filter=models.Filter(
                    must=[
                        models.FieldCondition(
                            key=p_version_hash, match=models.MatchValue(value=schema_hash)
                        )
                    ]
                ),
                limit=1,
            )
            if not response[0]:
                return None
            payload = response[0][0].payload
            return StorageSchemaInfo.from_normalized_mapping(payload, self.schema.naming)
        except UnexpectedResponse as e:
            if e.status_code == 404:
                return None
            raise
        except ValueError as e:  # Local qdrant error
            if "not found" in str(e):
                return None
            raise

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        return QDrantLoadJob(
            file_path,
            client_config=self.config,
            collection_name=self._make_qualified_collection_name(table["name"]),
        )

    def complete_load(self, load_id: str) -> None:
        values = [load_id, self.schema.name, 0, str(pendulum.now()), self.schema.version_hash]
        assert len(values) == len(self.loads_collection_properties)
        properties = {k: v for k, v in zip(self.loads_collection_properties, values)}
        loads_table_name = self._make_qualified_collection_name(self.schema.loads_table_name)
        self._create_point_no_vector(properties, loads_table_name)

    def __enter__(self) -> "QdrantClient":
        self.db_client = self.config.get_client()
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        if self.db_client:
            self.config.close_client(self.db_client)
            self.db_client = None

    def _update_schema_in_storage(self, schema: Schema) -> None:
        schema_str = json.dumps(schema.to_dict())
        values = [
            schema.version,
            schema.ENGINE_VERSION,
            str(pendulum.now().isoformat()),
            schema.name,
            schema.stored_version_hash,
            schema_str,
        ]
        assert len(values) == len(self.version_collection_properties)
        properties = {k: v for k, v in zip(self.version_collection_properties, values)}
        version_table_name = self._make_qualified_collection_name(self.schema.version_table_name)
        self._create_point_no_vector(properties, version_table_name)

    def _execute_schema_update(self, only_tables: Iterable[str]) -> None:
        is_local = self.config.is_local()
        for table_name in only_tables or self.schema.tables:
            exists = self._collection_exists(table_name)

            qualified_collection_name = self._make_qualified_collection_name(table_name)
            # NOTE: there are no property schemas in qdrant so we do not need to alter
            # existing collections
            if not exists:
                self._create_collection(
                    full_collection_name=qualified_collection_name,
                )
                if not is_local:  # Indexes don't work in local Qdrant (trigger log warning)
                    # Create indexes to enable order_by in state and schema tables
                    if table_name == self.schema.state_table_name:
                        self.db_client.create_payload_index(
                            collection_name=qualified_collection_name,
                            field_name=self.schema.naming.normalize_identifier("created_at"),
                            field_schema="datetime",
                        )
                    elif table_name == self.schema.version_table_name:
                        self.db_client.create_payload_index(
                            collection_name=qualified_collection_name,
                            field_name=self.schema.naming.normalize_identifier("inserted_at"),
                            field_schema="datetime",
                        )

        self._update_schema_in_storage(self.schema)

    def _collection_exists(self, table_name: str, qualify_table_name: bool = True) -> bool:
        try:
            table_name = (
                self._make_qualified_collection_name(table_name)
                if qualify_table_name
                else table_name
            )
            self.db_client.get_collection(table_name)
            return True
        except ValueError as e:
            if "not found" in str(e):
                return False
            raise
        except UnexpectedResponse as e:
            if e.status_code == 404:
                return False
            raise



import re
import base64
from typing import Optional

from dlt.common.exceptions import MissingDependencyException


PEM_REGEX = re.compile(
    r"-----BEGIN ([A-Z ]+)-----\s+([A-Za-z0-9+/=\s]+)-----END \1-----", re.MULTILINE
)


def is_pem(data: str) -> bool:
    return PEM_REGEX.search(data) is not None


def decode_private_key(private_key: str, password: Optional[str] = None) -> bytes:
    """Decode encrypted or unencrypted private key from string. Supported formats:
    1. base64 encoded DER
    2. plain-text or base64 encoded PEM
    """
    try:
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives.asymmetric import rsa
        from cryptography.hazmat.primitives.asymmetric import dsa
        from cryptography.hazmat.primitives import serialization
        from cryptography.hazmat.primitives.asymmetric.types import PrivateKeyTypes
    except ModuleNotFoundError as e:
        raise MissingDependencyException(
            "public / private key authentication",
            dependencies=["cryptography"],
        ) from e

    # check if this is PEM
    if is_pem(private_key):
        private_key_blob = private_key.encode(encoding="ascii")
    else:
        try:
            private_key_blob = base64.b64decode(private_key, validate=True)
        except Exception as der_exc:
            raise ValueError(
                "Could not decode private key for key pair authentication. Following formats were"
                f" attempted:\n1. base64 encoded DER or PEM (error: `{der_exc}`)\n2. plain-text PEM"
                " (error: BEGIN and END markers not found)\nIf you are using connection string and"
                " are passing DER/PEM or password in query string, make sure you url-encode them."
            )

    try:
        # load key as binary DER
        pkey = serialization.load_der_private_key(
            private_key_blob,
            password=password.encode() if password is not None else None,
            backend=default_backend(),
        )
    except Exception as der_exc:
        # loading DER key failed -> assume it's a plain-text PEM key
        try:
            pkey = serialization.load_pem_private_key(
                private_key_blob,
                password=password.encode() if password is not None else None,
                backend=default_backend(),
            )
        except Exception as pem_exc:
            raise ValueError(
                "Could not decode private key for key pair authentication. Following formats were"
                f" attempted:\n1. base64 encoded DER (error: `{der_exc}`)\n2. plain-text or base64"
                f" encoded PEM (error: `{pem_exc}`)\nIf you are using connection string and are"
                " passing DER/PEM or password in query string, make sure you url-encode them."
            )

    return pkey.private_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )



import warnings

from dlt.common.warnings import Dlt100DeprecationWarning


def uri_on_credentials_deprecated() -> None:
    warnings.warn(
        "Usage of `uri` argument on lance db credentials is deprecated. Please set the `lance_uri`"
        " on lance db config. for example using toml"
        ' section:\n[destination.lancedb]\nuri="path/db.lancedb"\n',
        Dlt100DeprecationWarning,
        stacklevel=1,
    )



"""
Custom Fabric Warehouse destination for dlt.

This module provides a custom destination class for Microsoft Fabric Warehouse
that extends Synapse capabilities with Fabric-specific requirements:
- Uses 'fabric' sqlglot dialect
- Uses varchar instead of nvarchar (Fabric doesn't support nvarchar)
- Supports COPY INTO for efficient data loading from staging
"""

from typing import Type, TYPE_CHECKING, Optional

from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.normalizers.naming import NamingConvention
from dlt.destinations.impl.synapse.factory import synapse, SynapseTypeMapper
from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.schema.typing import TColumnSchema

from .configuration import FabricClientConfiguration

if TYPE_CHECKING:
    from .fabric import FabricClient


class FabricTypeMapper(SynapseTypeMapper):
    """Custom type mapper for Fabric Warehouse - replaces nvarchar with varchar and datetimeoffset with datetime2"""

    def to_db_datetime_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema = None,
    ) -> str:
        """Override to limit precision to 6 and use datetime2 instead of datetimeoffset"""
        precision = column.get("precision")

        # Fabric only supports precision 0-6 (not 7 like SQL Server)
        if precision is not None:
            # Clamp to valid range for Fabric
            if precision < 0 or precision > 6:
                precision = 6

        # Fabric doesn't support datetimeoffset, always use datetime2
        # (timezone info is lost, but that's a Fabric limitation)
        if precision is not None:
            return f"datetime2({precision})"
        # Use precision 6 as default (instead of 7 in SQL Server)
        return "datetime2(6)"

    def to_destination_type(self, column: TColumnSchema, table: PreparedTableSchema) -> str:
        """Override to use varchar instead of nvarchar and datetime2 instead of datetimeoffset"""
        sc_t = column["data_type"]
        if sc_t == "json":
            # Fabric doesn't have native JSON type, use varchar instead of nvarchar
            return "varchar(%s)" % column.get("precision", "max")

        if sc_t == "time":
            # Fabric requires explicit precision for TIME (0-6)
            precision = min(column.get("precision", 6), 6)
            return "time(%i)" % precision

        # Get base type from parent
        db_type = super().to_destination_type(column, table)

        # Replace any nvarchar with varchar (Fabric doesn't support nvarchar)
        if "nvarchar" in db_type.lower():
            db_type = db_type.replace("nvarchar", "varchar").replace("NVARCHAR", "varchar")

        # Replace any datetimeoffset with datetime2 (Fabric doesn't support datetimeoffset)
        # This should be handled by to_db_datetime_type but just in case
        if "datetimeoffset" in db_type.lower():
            # Extract precision if present, e.g., datetimeoffset(7) -> datetime2(6)
            if "datetimeoffset(" in db_type.lower():
                import re

                match = re.search(r"datetimeoffset\((\d+)\)", db_type, re.IGNORECASE)
                if match:
                    precision = int(match.group(1))
                    precision = min(6, max(0, precision))  # Limit to 0-6
                    db_type = re.sub(
                        r"datetimeoffset\(\d+\)",
                        f"datetime2({precision})",
                        db_type,
                        flags=re.IGNORECASE,
                    )
                else:
                    db_type = db_type.replace("datetimeoffset", "datetime2").replace(
                        "DATETIMEOFFSET", "datetime2"
                    )
            else:
                db_type = db_type.replace("datetimeoffset", "datetime2").replace(
                    "DATETIMEOFFSET", "datetime2"
                )

        return db_type


class fabric(synapse):
    """Fabric Warehouse destination that extends Synapse with Fabric-specific dialect and type mappings"""

    spec: Type[FabricClientConfiguration] = FabricClientConfiguration  # type: ignore[assignment]

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        caps = super()._raw_capabilities()
        # Fabric uses the same file format strategy as Synapse:
        # - Direct loading: insert_values only (inherited from synapse)
        # - Staging: parquet (inherited from synapse)
        # Don't override preferred_loader_file_format or supported_loader_file_formats
        caps.has_case_sensitive_identifiers = True
        caps.sqlglot_dialect = "fabric"
        caps.type_mapper = FabricTypeMapper
        # Fabric only supports precision 0-6 for datetime2/time (not 7 like SQL Server)
        caps.max_timestamp_precision = 6
        return caps

    @classmethod
    def adjust_capabilities(
        cls,
        caps: DestinationCapabilitiesContext,
        config: FabricClientConfiguration,  # type: ignore[override]
        naming: Optional[NamingConvention],
    ) -> DestinationCapabilitiesContext:
        """Adjust capabilities based on collation case sensitivity"""
        # Modify caps if case sensitive identifiers are requested
        if config.has_case_sensitive_identifiers:
            caps.has_case_sensitive_identifiers = True
            caps.casefold_identifier = str
        return super().adjust_capabilities(caps, config, naming)  # type: ignore[arg-type]

    @property
    def client_class(self) -> Type["FabricClient"]:
        from .fabric import FabricClient

        return FabricClient


# Register the destination
fabric.register()



import base64
import string
import re
from typing import Sequence, Set, Optional, Any, Dict, Tuple

from dlt.common import logger
from dlt.common.json import json
from dlt.common.configuration.specs import GcpServiceAccountCredentials
from dlt.common.configuration.exceptions import ConfigProviderException
from dlt.common.exceptions import MissingDependencyException
from .vault import VaultDocProvider
from .provider import get_key_name

# Create a translation table to replace punctuation with ""
# since google secrets allow "-"" and "_" we need to exclude them
punctuation = "".join(set(string.punctuation) - {"-", "_"})
translator = str.maketrans("", "", punctuation)


def normalize_key(in_string: str) -> str:
    """Replaces punctuation characters in a string

    Note: We exclude `_` and `-` from punctuation characters

    Args:
        in_string(str): input string

    Returns:
        (str): a string without punctuation characters and whitespaces
    """

    # Strip punctuation from the string
    stripped_text = in_string.translate(translator)
    whitespace = re.compile(r"\s+")
    stripped_whitespace = whitespace.sub("", stripped_text)
    return stripped_whitespace


class GoogleSecretsProvider(VaultDocProvider):
    def __init__(
        self,
        credentials: GcpServiceAccountCredentials,
        only_secrets: bool = True,
        only_toml_fragments: bool = True,
        list_secrets: bool = False,
    ) -> None:
        """Initialize a Google Secrets Provider to access secrets stored in Google Secret Manager

        Args:
            credentials: Google Cloud credentials to access Secret Manager
            only_secrets: When True, only keys with secret hint types will be looked up
            only_toml_fragments: When True, only load known TOML fragments and ignore other lookups
            list_secrets: When True, list all secrets upfront to optimize vault access by
                          avoiding lookups for non-existent secrets. Requires additional
                          API calls and roles/secretmanager.secretViewer permission.
        """
        self.credentials = credentials
        super().__init__(only_secrets, only_toml_fragments, list_secrets)

    @staticmethod
    def get_key_name(key: str, *sections: str) -> str:
        """Make key name for the secret

        Per Google the secret name can contain, so we will use snake_case normalizer

            1. Uppercase and lowercase letters,
            2. Numerals,
            3. Hyphens,
            4. Underscores.
        """
        key = normalize_key(key)
        normalized_sections = [normalize_key(section) for section in sections if section]
        key_name = get_key_name(normalize_key(key), "-", *normalized_sections)
        return key_name

    @property
    def name(self) -> str:
        return "Google Secrets"

    @property
    def locations(self) -> Sequence[str]:
        if self.credentials:
            return [str(self.credentials)]
        else:
            return super().locations

    def _get_google_client(self, service_name: str = "secretmanager", version: str = "v1") -> Any:
        try:
            from googleapiclient.discovery import build
        except ModuleNotFoundError:
            raise MissingDependencyException(
                "GoogleSecretsProvider",
                ["google-api-python-client"],
                "We need google-api-python-client to build client for secretmanager v1",
            )

        return build(service_name, version, credentials=self.credentials.to_native_credentials())

    def _handle_http_error(self, error: Any) -> Tuple[Dict[str, Any], int, str, str]:
        error_doc = json.loadb(error.content)["error"]
        error_message = error_doc.get("message", "Unknown error")
        error_status = error_doc.get("status", "Unknown status")
        status_code = error.resp.status

        return error_doc, status_code, error_message, error_status

    def _look_vault(self, full_key: str, hint: type) -> Optional[str]:
        resource_name = f"projects/{self.credentials.project_id}/secrets/{full_key}/versions/latest"
        client = self._get_google_client()

        from googleapiclient.errors import HttpError

        try:
            response = client.projects().secrets().versions().access(name=resource_name).execute()
            secret_value = response["payload"]["data"]
            decoded_value = base64.b64decode(secret_value).decode("utf-8")
            return decoded_value
        except HttpError as error:
            _, status_code, error_message, error_status = self._handle_http_error(error)

            if status_code == 404:
                # logger.warning(f"{self.credentials.client_email} has roles/secretmanager.secretAccessor role but {full_key} not found in Google Secrets: {error_message}[{error_status}]")
                return None
            elif status_code == 403:
                logger.warning(
                    f"{self.credentials.client_email} does not have"
                    " roles/secretmanager.secretAccessor role. It also does not have read"
                    f" permission to {full_key} or the key is not found in Google Secrets:"
                    f" {error_message}[{error_status}]"
                )
                return None
            elif status_code == 400:
                logger.warning(f"Unable to read {full_key} : {error_message}[{error_status}]")
                return None
            raise

    def _list_vault(self) -> Set[str]:
        """Lists all available secrets in Google Secret Manager

        This method is called when list_secrets is enabled to prefetch all available
        secret names, which helps avoid unnecessary API calls for non-existent secrets.

        Returns:
            Set[str]: A set of available key names in the vault
        """
        available_keys: Set[str] = set()
        client = self._get_google_client()

        from googleapiclient.errors import HttpError

        try:
            parent = f"projects/{self.credentials.project_id}"
            request = client.projects().secrets().list(parent=parent)

            while request is not None:
                response = request.execute()
                secrets = response.get("secrets", [])

                for secret in secrets:
                    # Extract secret name from full resource path
                    name = secret.get("name", "").split("/")[-1]
                    if name:
                        available_keys.add(name)

                request = client.projects().secrets().list_next(request, response)

            logger.info(f"Listed {len(available_keys)} secrets from Google Secret Manager")
            return available_keys

        except HttpError as error:
            _, status_code, error_message, error_status = self._handle_http_error(error)

            if status_code == 403:
                raise ConfigProviderException(
                    self.name,
                    f"Cannot list secrets: `{self.credentials.client_email}` does not have "
                    "roles/secretmanager.secretViewer role. Secret listing is required when "
                    "list_secrets=True to optimize vault access by skipping lookups for "
                    f"non-existent secrets. Error: {error_message} [{error_status}]",
                )
            else:
                raise ConfigProviderException(
                    self.name,
                    "Failed to list secrets in Google Secret Manager. Secret listing is required"
                    " when list_secrets=True to optimize vault access by skipping lookups for"
                    f" non-existent secrets. Error: {error_message} [{error_status}]",
                )



from __future__ import annotations

import pathlib
import textwrap
from typing import TYPE_CHECKING, Any

from dlt.common.storages import FileStorage
from dlt.common.pipeline import get_dlt_pipelines_dir

if TYPE_CHECKING:
    import pyarrow


def _load_pickle(file_path: pathlib.Path) -> Any:
    import pickle

    return pickle.loads(file_path.read_bytes())


def _load_json(file_path: pathlib.Path) -> Any:
    from dlt.common import json

    return json.loads(file_path.read_text())


def _load_jsonl(file_path: pathlib.Path) -> list[Any]:
    from dlt.common import json

    with file_path.open("r") as f:
        records = [json.loads(line.strip("\n")) for line in f.readlines()]
    return records


def _load_gzip(file_path: pathlib.Path) -> Any:
    return FileStorage.open_zipsafe_ro(str(file_path)).read()


def _load_parquet(file_path: pathlib.Path) -> pyarrow.Table:
    import pyarrow.parquet

    return pyarrow.parquet.read_table(file_path)


def _load_csv(file_path: pathlib.Path) -> pyarrow.Table:
    import pyarrow.csv

    return pyarrow.csv.read_csv(file_path)


def _load_insert_values_gzip(file_path: pathlib.Path) -> pyarrow.Table:
    import sqlglot
    import duckdb

    table_name = file_path.name.partition(".")[0]

    insert_statement = _load_gzip(file_path)
    # Remove the `E` prefix used by postgres-style quote escapes
    insert_statement = insert_statement.replace("E'", r"'")
    # For some reason, the INSERT statement is missing the table name
    # we can get this info from the file name
    insert_statement = insert_statement.replace("{}", f"{table_name} ")
    insert_expr = sqlglot.parse(insert_statement)[0]

    query = textwrap.dedent(f"""\
        CREATE TABLE {table_name} AS
        FROM ({insert_expr.expression.sql()})
        AS {insert_expr.this.sql()}
        """)

    con = duckdb.connect(":memory:")
    con.execute(query)
    arrow_table = con.execute(f"FROM {table_name}").fetch_arrow_table()
    con.close()  # release the memory
    return arrow_table


def _load_raw_text(file_path: pathlib.Path) -> str:
    return file_path.read_text()


def _load_raw_bytes(file_path: pathlib.Path) -> bytes:
    return file_path.read_bytes()


def _load_file(file_path: pathlib.Path) -> Any:
    try:
        if file_path.suffix == ".pickle":
            return _load_pickle(file_path)
        elif file_path.suffix == ".json":
            return _load_json(file_path)
        elif file_path.suffix == ".jsonl":
            return _load_jsonl(file_path)
        elif ".insert_values" in file_path.suffixes and file_path.suffix in (".gzip", ".gz"):
            return _load_insert_values_gzip(file_path)
        elif file_path.suffix in (".gzip", ".gz"):
            return _load_gzip(file_path)
        elif file_path.suffix == ".parquet":
            return _load_parquet(file_path)
        elif file_path.suffix == ".csv":
            return _load_csv(file_path)
        else:
            return _load_raw_text(file_path)
    except Exception:
        return _load_raw_bytes(file_path)



import builtins
import importlib.util
from importlib import import_module
from types import ModuleType, SimpleNamespace
from typing import Any, Callable, Literal, NamedTuple, Tuple, Mapping, List, Sequence

from dlt.common.exceptions import MissingDependencyException, TypeErrorWithKnownTypes
from dlt.common.typing import TAny


class DummyModule(ModuleType):
    """A dummy module from which you can import anything"""

    def __getattr__(self, key: str) -> Any:
        if key[0].isupper():
            # if imported name is capitalized, import type
            return SimpleNamespace
        else:
            # otherwise import instance
            return SimpleNamespace()

    __all__: List[Any] = []  # support wildcard imports


def import_module_with_missing(name: str, missing_modules: Tuple[str, ...] = ()) -> ModuleType:
    """Module importer that ignores missing modules by importing a dummy module"""

    def _try_import(
        name: str,
        _globals: Mapping[str, Any] = None,
        _locals: Mapping[str, Any] = None,
        fromlist: Sequence[str] = (),
        level: int = 0,
    ) -> ModuleType:
        """This function works as follows: on ImportError it raises. This import error is then next caught in the main function body and the name is added to exceptions.
        Next time if the name is on exception list or name is a package on exception list we return DummyModule and do not reraise
        This excepts only the modules that bubble up ImportError up until our code so any handled import errors are not excepted
        """
        try:
            return real_import(name, _globals, _locals, fromlist, level)
        except ImportError:
            # print(f"_import_module {name} {missing_modules} {fromlist} {level} {ex}")
            # return a dummy when: (1) name is on exception list (2) name is package path (dot separated) that start with exception from the list
            if any(name == m or name.startswith(m + ".") for m in missing_modules):
                return DummyModule(name)
            else:
                raise

    try:
        # patch built in import
        real_import, builtins.__import__ = builtins.__import__, _try_import  # type: ignore
        # discover missing modules and repeat until all are patched by dummies
        while True:
            try:
                return import_module(name)
            # `MissingDependencyException` is an `ImportError`
            except MissingDependencyException as me:
                if isinstance(me.__context__, ImportError):
                    if me.__context__.name is None:
                        raise
                    if me.__context__.name in missing_modules:
                        # print(f"{me.__context__.name} IN :/")
                        raise
                    # print(f"ADD {me.__context__.name}")
                    missing_modules += (me.__context__.name,)
                else:
                    raise
            # catch `ImportError` that are not `MissingDependencyException`
            except ImportError as ie:
                if ie.name is None:
                    raise
                # print(f"ADD {ie.name} {ie.path} vs {name} vs {str(ie)}")
                if ie.name in missing_modules:
                    raise
                missing_modules += (ie.name,)
    finally:
        builtins.__import__ = real_import


class ImportTrace(NamedTuple):
    ref: str
    module: str
    attr_name: str
    reason: Literal[
        "MissingDependencyException",
        "ModuleSpecNotFound",
        "ImportSpecError",
        "ImportError",
        "AttrNotFound",
        "TypeCheck",
    ]
    exc: Exception


def callable_typechecker(o: TAny) -> TAny:
    if callable(o):
        return o  # type: ignore[no-any-return]
    raise TypeErrorWithKnownTypes("attr", o, ["Callable"])


def object_from_ref(
    ref: str,
    typechecker: Callable[[Any], Any],
    raise_exec_errors: bool = False,
    import_missing_modules: bool = False,
) -> Tuple[Any, ImportTrace]:
    """Splits ref on module.attr and import module, then gets attr and runs typechecker on it.
    If `import_missing_modules` is True, will use custom importer that replaces missing modules with dummy
    from which you can import anything.
    NOTE: if your import got patched it will not work. Use this only to import refs ie. to verify if they exist, not in production code
    if `raise_exec_errors`, all errors that happen when code of the imported module is executed are raised, otherwise corresponding trace is returned

    Returns: a tuple (typechecked attr, trace if error)

    """
    if "." not in ref:
        raise ValueError("`ref` format is `module.attr` and must contain at least one `.`")
    module_path, attr_name = ref.rsplit(".", 1)
    try:
        spec = importlib.util.find_spec(module_path)
        if spec is None:
            return None, ImportTrace(ref, module_path, attr_name, "ModuleSpecNotFound", None)
    except Exception as exc:
        # one of modules on path could not be imported
        if isinstance(exc, ModuleNotFoundError) and module_path.startswith(exc.name):
            return None, ImportTrace(ref, module_path, attr_name, "ModuleSpecNotFound", exc)
        if raise_exec_errors:
            raise
        # spec may raise when importing parent packages of the module_path
        return None, ImportTrace(ref, module_path, attr_name, "ImportSpecError", exc)

    _importer = import_module_with_missing if import_missing_modules else import_module

    try:
        dest_module = _importer(module_path)
    except MissingDependencyException as missing_ex:
        if raise_exec_errors:
            raise
        return None, ImportTrace(
            ref, module_path, attr_name, "MissingDependencyException", missing_ex
        )
    except Exception as mod_ex:
        if raise_exec_errors:
            raise
        return None, ImportTrace(ref, module_path, attr_name, "ImportError", mod_ex)

    try:
        factory = getattr(dest_module, attr_name)
    except AttributeError as attr_ex:
        return None, ImportTrace(ref, module_path, attr_name, "AttrNotFound", attr_ex)

    # check type
    try:
        return typechecker(factory), None
    except Exception as type_exc:
        return None, ImportTrace(ref, module_path, attr_name, "TypeCheck", type_exc)



import warnings
from dlt.common.warnings import Dlt100DeprecationWarning


def per_column_cluster_hint_deprecated(clustered_columns: list[str]) -> None:
    columns_str = ", ".join(f"'{col}'" for col in clustered_columns)
    example = (
        "\n\nTo migrate, use the 'cluster' parameter in the `bigquery_adapter` function, for"
        f" example:\n    bigquery_adapter(resource, cluster=[{columns_str}])\nThis ensures the"
        " order of clustered columns is preserved.\n"
    )
    warnings.warn(
        "Defining clustered tables in BigQuery using per-column 'cluster' hints is deprecated and"
        " will be removed in a future release. Clustered columns detected:"
        f" [{columns_str}].{example}",
        Dlt100DeprecationWarning,
        stacklevel=2,
    )



from typing import Literal, Dict, Set, Union

from dlt.common.schema import TColumnHint
from dlt.common.typing import get_args


TSecureConnection = Literal[0, 1]
TTableEngineType = Literal[
    "merge_tree",
    "shared_merge_tree",
    "replicated_merge_tree",
]
TColumnName = str
TSQLExprOrColumnSeq = Union[str, list[TColumnName], tuple[TColumnName]]
TMergeTreeSettingsValue = Union[str, int, float, bool]
TMergeTreeSettings = Dict[str, TMergeTreeSettingsValue]
TCodecExpr = str
TColumnCodecs = Dict[TColumnName, TCodecExpr]

HINT_TO_CLICKHOUSE_ATTR: Dict[TColumnHint, str] = {
    "primary_key": "PRIMARY KEY",
    "unique": "",  # No unique constraints available in ClickHouse.
}

TABLE_ENGINE_TYPE_TO_CLICKHOUSE_ATTR: Dict[TTableEngineType, str] = {
    "merge_tree": "MergeTree",
    "shared_merge_tree": "SharedMergeTree",
    "replicated_merge_tree": "ReplicatedMergeTree",
}

TDeployment = Literal["ClickHouseOSS", "ClickHouseCloud"]

SUPPORTED_FILE_FORMATS = Literal["jsonl", "parquet"]
FILE_FORMAT_TO_TABLE_FUNCTION_MAPPING: Dict[SUPPORTED_FILE_FORMATS, str] = {
    "jsonl": "JSONEachRow",
    "parquet": "Parquet",
}
TABLE_ENGINE_TYPES: Set[TTableEngineType] = set(get_args(TTableEngineType))
TABLE_ENGINE_TYPE_HINT: Literal["x-table-engine-type"] = "x-table-engine-type"
SORT_HINT: Literal["x-clickhouse-sort"] = "x-clickhouse-sort"
PARTITION_HINT: Literal["x-clickhouse-partition"] = "x-clickhouse-partition"
SETTINGS_HINT: Literal["x-clickhouse-settings"] = "x-clickhouse-settings"
CODEC_HINT: Literal["x-clickhouse-codec"] = "x-clickhouse-codec"



from typing import Iterator, List

from dlt.common.destination.client import (
    PreparedTableSchema,
    RunnableLoadJob,
    HasFollowupJobs,
    LoadJob,
)
from dlt.common.storages import FileStorage
from dlt.common.storages.load_storage import ParsedLoadJobFileName
from dlt.common.utils import chunks

from dlt.destinations.job_client_impl import SqlJobClientWithStagingDataset, SqlJobClientBase


class InsertValuesLoadJob(RunnableLoadJob, HasFollowupJobs):
    def __init__(self, file_path: str) -> None:
        super().__init__(file_path)
        self._job_client: "SqlJobClientBase" = None

    def run(self) -> None:
        # insert file content immediately
        self._sql_client = self._job_client.sql_client

        with self._sql_client.begin_transaction():
            for fragments in self._insert(
                self._sql_client.make_qualified_table_name(self.load_table_name), self._file_path
            ):
                self._sql_client.execute_fragments(fragments)

    def _insert(self, qualified_table_name: str, file_path: str) -> Iterator[List[str]]:
        # WARNING: maximum redshift statement is 16MB https://docs.aws.amazon.com/redshift/latest/dg/c_redshift-sql.html
        # the procedure below will split the inserts into max_query_length // 2 packs
        with FileStorage.open_zipsafe_ro(file_path, "r", encoding="utf-8") as f:
            header = f.readline()
            # format and casefold header
            header = self._sql_client.capabilities.casefold_identifier(header).format(
                qualified_table_name
            )
            writer_type = self._sql_client.capabilities.insert_values_writer_type
            if writer_type == "default":
                sep = ","
                # properly formatted file has a values marker at the beginning
                values_mark = f.readline()
                assert values_mark == "VALUES\n"
            elif writer_type == "select_union":
                sep = " UNION ALL"

            max_rows = self._sql_client.capabilities.max_rows_per_insert

            insert_sql = []
            while content := f.read(self._sql_client.capabilities.max_query_length // 2):
                # read one more line in order to
                # 1. complete the content which ends at "random" position, not an end line
                # 2. to modify its ending without a need to re-allocating the 8MB of "content"
                until_nl = f.readline()
                # if until next line contains just '\n' try to take another line so we can finish content properly
                # TODO: write test for this case (content ends with ",")
                if until_nl == "\n":
                    until_nl = f.readline()
                until_nl = until_nl.strip("\n")
                # if there was anything left, until_nl contains the last line
                is_eof = len(until_nl) == 0 or until_nl[-1] == ";"
                if not is_eof:
                    until_nl = until_nl[: -len(sep)] + ";"  # replace the separator with ";"
                if max_rows is not None:
                    # mssql has a limit of 1000 rows per INSERT, so we need to split into separate statements
                    values_rows = content.splitlines(keepends=True)
                    len_rows = len(values_rows)
                    processed = 0
                    # Chunk by max_rows - 1 for simplicity because one more row may be added
                    for chunk in chunks(values_rows, max_rows - 1):
                        processed += len(chunk)
                        insert_sql.append(header)
                        if writer_type == "default":
                            insert_sql.append(values_mark)
                        if processed == len_rows:
                            # On the last chunk we need to add the extra row read
                            insert_sql.append("".join(chunk) + until_nl)
                        else:
                            # Replace the , with ;
                            insert_sql.append("".join(chunk).strip()[: -len(sep)] + ";\n")
                else:
                    # otherwise write all content in a single INSERT INTO
                    if writer_type == "default":
                        insert_sql.extend([header, values_mark, content + until_nl])
                    elif writer_type == "select_union":
                        insert_sql.extend([header, content + until_nl])

                # actually this may be empty if we were able to read a full file into content
                if not is_eof:
                    # execute chunk of insert
                    yield insert_sql
                    insert_sql = []

        if insert_sql:
            yield insert_sql


class InsertValuesJobClient(SqlJobClientWithStagingDataset):
    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        job = super().create_load_job(table, file_path, load_id, restore)
        if not job:
            parsed_file = ParsedLoadJobFileName.parse(file_path)
            # this is using sql_client internally and will raise a right exception
            if parsed_file.file_format == "insert_values":
                job = InsertValuesLoadJob(file_path)
        return job



"""
  Copyright (C) 2017-2021 Dremio Corporation

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

      http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

The code in this module was original from https://github.com/dremio-hub/arrow-flight-client-examples/tree/main/python.
The code has been modified and extended to provide a PEP 249 compatible interface.

This implementation will eagerly gather the full result set after every query.
Eventually, this module should be replaced with ADBC Flight SQL client.
See: https://github.com/apache/arrow-adbc/issues/1559
"""

from dataclasses import dataclass, field
from datetime import datetime  # noqa: I251
from http.cookies import SimpleCookie
from typing import Any, List, Tuple, Optional, Mapping, Dict, AnyStr, Iterator

import pyarrow
import pytz
from pyarrow import flight

apilevel = "2.0"
threadsafety = 2
paramstyle = "format"


def connect(
    uri: str,
    db_kwargs: Optional[Mapping[str, Any]] = None,
    conn_kwargs: Optional[Mapping[str, Any]] = None,
) -> "DremioConnection":
    username = db_kwargs["username"]
    password = db_kwargs["password"]
    tls_root_certs = db_kwargs.get("tls_root_certs")
    client = create_flight_client(location=uri, tls_root_certs=tls_root_certs)
    options = create_flight_call_options(
        username=username,
        password=password,
        client=client,
    )
    return DremioConnection(
        client=client,
        options=options,
    )


def quote_string(string: str) -> str:
    return "'" + string.strip("'") + "'"


def format_datetime(d: datetime) -> str:
    if d.tzinfo is not None:
        d = d.astimezone(pytz.UTC).replace(tzinfo=None)
    return d.isoformat(sep=" ", timespec="milliseconds")


def format_parameter(param: Any) -> str:
    if isinstance(param, str):
        return quote_string(param)
    elif isinstance(param, datetime):
        return quote_string(format_datetime(param))
    else:
        return str(param)


class MalformedQueryError(Exception):
    pass


def parameterize_query(query: str, parameters: Optional[Tuple[Any, ...]]) -> str:
    parameters = parameters or ()
    parameters = tuple(format_parameter(p) for p in parameters)
    try:
        return query % parameters
    except TypeError as ex:
        raise MalformedQueryError(*ex.args)


def _any_str_to_str(string: AnyStr) -> str:
    if isinstance(string, bytes):
        return string.decode()
    else:
        return string


@dataclass
class DremioCursor:
    connection: "DremioConnection"
    table: pyarrow.Table = field(init=False, default_factory=lambda: pyarrow.table([]))
    # default number of rows returned by fetchmany when size is not provided
    arraysize: int = 1000
    # flight stream reader and schema for streaming fetches
    _reader: Optional[flight.FlightStreamReader] = field(init=False, default=None)
    _schema: Optional[pyarrow.Schema] = field(init=False, default=None)

    @property
    def description(self) -> List[Tuple[str, pyarrow.DataType, Any, Any, Any, Any, Any]]:
        schema = self._schema or self.table.schema
        return [(fld.name, fld.type, None, None, None, None, None) for fld in schema]

    @property
    def rowcount(self) -> int:
        return len(self.table)

    def execute(
        self, query: AnyStr, parameters: Optional[Tuple[Any, ...]] = None, *args: Any, **kwargs: Any
    ) -> None:
        """Execute a query and prepare to stream results.

        this configures a Flight stream reader so that results can be consumed
        incrementally via fetchone/fetchmany/fetchall or iter_arrow_tables.

        Args:
            query: SQL query string (or bytes)
            parameters: optional positional parameters for the query
        """
        # cancel any prior reader
        if self._reader is not None:
            try:
                self._reader.cancel()
            except Exception:
                pass
            finally:
                self._reader = None

        query_str = _any_str_to_str(query)
        parameterized_query = parameterize_query(query_str, parameters)

        # get flight info and open a reader without loading all results
        flight_desc = flight.FlightDescriptor.for_command(parameterized_query)
        flight_info = self.connection.client.get_flight_info(flight_desc, self.connection.options)
        self._schema = flight_info.schema
        self._reader = self.connection.client.do_get(
            flight_info.endpoints[0].ticket, self.connection.options
        )

        # reset buffer table to empty with the expected schema
        if self._schema is not None:
            self.table = pyarrow.table({col.name: [] for col in self._schema}, schema=self._schema)
        else:
            self.table = pyarrow.table([])

    def fetchall(self) -> List[Tuple[Any, ...]]:
        # read all remaining rows in batches of arraysize to avoid high memory spikes
        rows: List[Tuple[Any, ...]] = []
        while True:
            batch = self.fetchmany(self.arraysize)
            if not batch:
                break
            rows.extend(batch)
        return rows

    def _ensure_buffered_rows(self, size: int) -> None:
        """ensure the in-memory buffer has at least size rows, reading from the stream.

        Args:
            size: minimum number of rows to buffer in self.table.
        """
        while len(self.table) < size and self._reader is not None:
            try:
                chunk = self._reader.read_chunk()
            except StopIteration:
                self._reader = None
                break
            record_batch = chunk.data
            if record_batch is None:
                continue
            incoming = pyarrow.Table.from_batches([record_batch])
            self.table = (
                incoming if len(self.table) == 0 else pyarrow.concat_tables([self.table, incoming])
            )

    def fetchmany(self, size: Optional[int] = None) -> List[Tuple[Any, ...]]:
        """Fetch up to size rows, streaming from the server as needed.

        Args:
            size: number of rows to return. If None, uses self.arraysize.

        Returns:
            list of row tuples with up to ``size`` elements. May be empty when
            no more rows are available.
        """
        if size is None:
            size = self.arraysize
        if size <= 0:
            return []

        # ensure at least size rows in buffer
        self._ensure_buffered_rows(size)

        available = len(self.table)
        if available == 0:
            return []

        take = min(size, available)
        slice_table = self.table.slice(0, take)
        self.table = self.table.slice(take)
        return [tuple(d.values()) for d in slice_table.to_pylist()]

    def fetchone(self) -> Optional[Tuple[Any, ...]]:
        result = self.fetchmany(1)
        return result[0] if result else None

    def fetch_arrow_table(self) -> pyarrow.Table:
        """Read and return the full remaining result as an Arrow table.

        this drains the underlying Flight stream (if active) and returns all
        rows currently buffered plus all remaining rows from the server. The
        cursor buffer is reset to empty afterwards.
        """
        parts: List[pyarrow.Table] = []
        if len(self.table) > 0:
            parts.append(self.table)

        if self._reader is not None:
            try:
                while True:
                    try:
                        chunk = self._reader.read_chunk()
                    except StopIteration:
                        self._reader = None
                        break
                    rb = chunk.data
                    if rb is None:
                        continue
                    parts.append(pyarrow.Table.from_batches([rb]))
            finally:
                # guard against leaving a half-open reader reference
                self._reader = None

        if parts:
            full = pyarrow.concat_tables(parts) if len(parts) > 1 else parts[0]
        else:
            # return an empty table preserving known schema when possible
            schema = self._schema or self.table.schema
            full = pyarrow.table({col.name: [] for col in schema}, schema=schema)

        # reset buffer to empty with the same schema
        self.table = pyarrow.table({col.name: [] for col in full.schema}, schema=full.schema)
        return full

    def close(self) -> None:
        # cancel the active reader, if any
        if self._reader is not None:
            try:
                self._reader.cancel()
            except Exception:
                pass
            finally:
                self._reader = None

    def __enter__(self) -> "DremioCursor":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:  # type: ignore
        self.close()

    def iter_arrow_tables(self, size: Optional[int] = None) -> pyarrow.Table:
        """Fetch up to size rows and return them as an Arrow table.

        this mirrors fetchmany but returns a pyarrow.Table instead of a list of
        row tuples. It streams from the server as needed to satisfy the size.

        Args:
            size: number of rows to return. If None, uses self.arraysize.

        Returns:
            pyarrow.Table containing up to ``size`` rows. May be empty when no
            more rows are available.
        """
        if size is None:
            size = self.arraysize
        if size <= 0:
            # return an empty table with the current schema
            schema = self.table.schema
            return pyarrow.table({col.name: [] for col in schema}, schema=schema)

        # ensure at least size rows in buffer
        self._ensure_buffered_rows(size)

        take = min(size, len(self.table))
        if take == 0:
            return None

        slice_table = self.table.slice(0, take)
        self.table = self.table.slice(take)
        return slice_table


@dataclass(frozen=True)
class DremioConnection:
    client: flight.FlightClient
    options: flight.FlightCallOptions

    def close(self) -> None:
        self.client.close()

    def cursor(self) -> DremioCursor:
        return DremioCursor(self)

    def __enter__(self) -> "DremioConnection":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:  # type: ignore
        self.close()


class DremioAuthError(Exception):
    pass


class DremioClientAuthMiddlewareFactory(flight.ClientMiddlewareFactory):
    """A factory that creates DremioClientAuthMiddleware(s)."""

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.call_credential: Optional[Tuple[bytes, bytes]] = None

    def start_call(self, info: flight.CallInfo) -> flight.ClientMiddleware:
        return DremioClientAuthMiddleware(self)

    def set_call_credential(self, call_credential: Tuple[bytes, bytes]) -> None:
        self.call_credential = call_credential


class DremioClientAuthMiddleware(flight.ClientMiddleware):
    """
    A ClientMiddleware that extracts the bearer token from
    the authorization header returned by the Dremio
    Flight Server Endpoint.

    Parameters
    ----------
    factory : ClientHeaderAuthMiddlewareFactory
        The factory to set call credentials if an
        authorization header with bearer token is
        returned by the Dremio server.
    """

    def __init__(self, factory: flight.ClientMiddlewareFactory, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.factory = factory

    def received_headers(self, headers: Mapping[str, str]) -> None:
        auth_header_key = "authorization"
        authorization_header = None
        for key in headers:
            if key.lower() == auth_header_key:
                authorization_header = headers.get(auth_header_key)
        if authorization_header:
            self.factory.set_call_credential(
                (b"authorization", authorization_header[0].encode("utf-8"))
            )


class CookieMiddlewareFactory(flight.ClientMiddlewareFactory):
    """A factory that creates CookieMiddleware(s)."""

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.cookies: Dict[str, Any] = {}

    def start_call(self, info: flight.CallInfo) -> flight.ClientMiddleware:
        return CookieMiddleware(self)


class CookieMiddleware(flight.ClientMiddleware):
    """
    A ClientMiddleware that receives and retransmits cookies.
    For simplicity, this does not auto-expire cookies.

    Parameters
    ----------
    factory : CookieMiddlewareFactory
        The factory containing the currently cached cookies.
    """

    def __init__(self, factory: CookieMiddlewareFactory, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.factory = factory

    def received_headers(self, headers: Mapping[str, str]) -> None:
        for key in headers:
            if key.lower() == "set-cookie":
                cookie = SimpleCookie()
                for item in headers.get(key):
                    cookie.load(item)

                self.factory.cookies.update(cookie.items())

    def sending_headers(self) -> Dict[bytes, bytes]:
        if self.factory.cookies:
            cookie_string = "; ".join(
                "{!s}={!s}".format(key, val.value) for (key, val) in self.factory.cookies.items()
            )
            return {b"cookie": cookie_string.encode("utf-8")}
        return {}


# def tls_root_certs() -> bytes:
#     with open("certs/ca-certificates.crt", "rb") as f:
#         return f.read()


def create_flight_client(
    location: str, tls_root_certs: Optional[bytes] = None, **kwargs: Any
) -> flight.FlightClient:
    return flight.FlightClient(
        location=location,
        tls_root_certs=tls_root_certs,
        middleware=[DremioClientAuthMiddlewareFactory(), CookieMiddlewareFactory()],
        **kwargs,
    )


def create_flight_call_options(
    username: str, password: str, client: flight.FlightClient
) -> flight.FlightCallOptions:
    headers: List[Any] = []
    # Retrieve bearer token and append to the header for future calls.
    bearer_token = client.authenticate_basic_token(
        username,
        password,
        flight.FlightCallOptions(headers=headers),
    )
    headers.append(bearer_token)
    return flight.FlightCallOptions(headers=headers)



from __future__ import annotations
import time
import multiprocessing
from typing import Any, Callable, Optional, Tuple, Union, cast, TypeVar
from concurrent.futures import Executor, ThreadPoolExecutor, Future
from typing_extensions import ParamSpec

from dlt.common import logger
from dlt.common.configuration.container import Container
from dlt.common.configuration.specs.pluggable_run_context import PluggableRunContext
from dlt.common.runtime import init
from dlt.common.runners.runnable import Runnable, TExecutor
from dlt.common.runners.configuration import PoolRunnerConfiguration
from dlt.common.runners.typing import TRunMetrics
from dlt.common.runtime import signals
from dlt.common.runtime.signals import sleep
from dlt.common.exceptions import SignalReceivedException
from dlt.common.typing import AnyFun
from dlt.common.runtime.exec_info import platform_supports_threading

T = TypeVar("T")
P = ParamSpec("P")


class NullExecutor(Executor):
    """Dummy executor that runs jobs single-threaded.

    Provides a uniform interface for `None` pool type
    """

    def submit(self, fn: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs) -> Future[T]:
        """Run the job and return a Future"""
        fut: Future[T] = Future()
        try:
            result = fn(*args, **kwargs)
        except BaseException as exc:
            fut.set_exception(exc)
        else:
            fut.set_result(result)
        return fut


class TimeoutThreadPoolExecutor(ThreadPoolExecutor):
    def __init__(
        self,
        max_workers: int = None,
        timeout: Optional[float] = None,
        thread_name_prefix: str = "",
        initializer: AnyFun = None,
        initargs: Tuple[Any, ...] = (),
    ):
        """Initializes a new ThreadPoolExecutor instance.

        Args:
            max_workers: The maximum number of threads that can be used to
                execute the given calls.
            timeout: Waits for pool threads to complete, None to wait indefinite
            thread_name_prefix: An optional name prefix to give our threads.
            initializer: A callable used to initialize worker threads.
            initargs: A tuple of arguments to pass to the initializer.
        """
        super().__init__(
            max_workers,
            thread_name_prefix=thread_name_prefix,
            initializer=initializer,
            initargs=initargs,
        )
        self.timeout = timeout
        self._is_alive: bool = None  # set on shutdown

    def shutdown(self, wait: bool = True, *, cancel_futures: bool = False) -> None:
        if not wait or self.timeout is None:
            # behave exactly like the std-lib version
            return super().shutdown(wait=wait, cancel_futures=cancel_futures)

        # initiate shutdown but return immediately
        super().shutdown(wait=False, cancel_futures=cancel_futures)

        deadline = time.monotonic() + self.timeout
        for t in list(self._threads):
            remaining = deadline - time.monotonic()
            if remaining <= 0:
                break
            t.join(remaining)

        alive = [t for t in self._threads if t.is_alive()]
        if alive:
            logger.error(
                f"Shutdown of dlt thread pool {self._thread_name_prefix} could not be completed in"
                f" {self.timeout} seconds. All the tasks in the pool are completed, nevertheless"
                " some threads didn't join. This may happen, for example, if thread local storage "
                " is used with object with __del__ method that locks thread when garbage"
                " collecting or if os.fork() is used. NOTE: if you use database clients or other"
                " Python libraries in resource or custom destinations, they can do that without"
                " you knowing. WARNING! This process will still try to join those threads on exit"
                " and most probably will lock indefinitely."
            )
            logger.info(
                f"Shutdown of dlt thread pool {self._thread_name_prefix} - "
                f"{len(alive)} thread(s) still alive: {', '.join(t.name for t in alive)}."
            )
        self._is_alive = len(alive) > 0


def get_default_start_method(method_: str) -> str:
    """Sets method to `spawn` is running in one of orchestrator tasks.

    Called when explicit start method is not set on `PoolRunnerConfiguration`
    """
    if method_ == "fork":
        from dlt.common.runtime.exec_info import (
            is_running_in_airflow_task,
            is_running_in_dagster_task,
            is_running_in_prefect_flow,
        )

        for m_ in [
            is_running_in_airflow_task,
            is_running_in_dagster_task,
            is_running_in_prefect_flow,
        ]:
            if m_():
                logger.info(
                    f"Switching pool start method to `spawn` because `{m_.__name__}` is True"
                )
                return "spawn"
    return method_


def create_pool(config: PoolRunnerConfiguration) -> Executor:
    assert config.pool_type in ["process", "thread", "none", None]

    if not platform_supports_threading():
        logger.info(
            "Platform does not support threading, created single-threaded pool for execution."
        )
        return NullExecutor()

    executor: Executor = None
    if config.pool_type == "process":
        # import process pool only when needed. not all Python envs support process pools
        from concurrent.futures import ProcessPoolExecutor

        # if not fork method, provide initializer for logs and configuration
        start_method = config.start_method or get_default_start_method(
            multiprocessing.get_start_method()
        )
        # get the mp_context for the configured start method
        mp_context = multiprocessing.get_context(method=start_method)
        if start_method != "fork":
            container = Container()
            run_context = container[PluggableRunContext].context
            worker_contexts = container.get_worker_contexts()

            executor = ProcessPoolExecutor(
                max_workers=config.workers,
                initializer=init.restore_run_context,
                initargs=(run_context, worker_contexts),
                mp_context=mp_context,
            )
        else:
            executor = ProcessPoolExecutor(max_workers=config.workers, mp_context=mp_context)
    elif config.pool_type == "thread":
        # wait 2 seconds before bailing out from pool shutdown
        executor = TimeoutThreadPoolExecutor(
            max_workers=config.workers,
            timeout=2.0,
            thread_name_prefix=Container.thread_pool_prefix(),
        )
    # no pool - single threaded
    else:
        executor = NullExecutor()

    logger.info(
        f"Created {config.pool_type or 'single-threaded'} pool with"
        f" {config.workers or 'default no.'} workers"
    )
    return executor


def run_pool(
    config: PoolRunnerConfiguration,
    run_f: Union[Runnable[TExecutor], Callable[[TExecutor], TRunMetrics]],
) -> int:
    # validate the run function
    if not isinstance(run_f, Runnable) and not callable(run_f):
        raise ValueError(
            run_f, "Pool runner entry point must be a function `f(pool: TPool)` or `Runnable`"
        )

    # start pool
    pool = create_pool(config)
    runs_count = 1

    def _run_func() -> bool:
        if callable(run_f):
            run_metrics = run_f(cast(TExecutor, pool))
        elif isinstance(run_f, Runnable):
            run_metrics = run_f.run(cast(TExecutor, pool))
        else:
            raise SignalReceivedException(-1)
        return run_metrics.pending_items > 0

    try:
        logger.debug("Running pool")
        while _run_func():
            # raise on signal: safe to do that out of _run_func()
            signals.raise_if_signalled()
            runs_count += 1
            sleep(config.run_sleep)
            # signal could come
            signals.raise_if_signalled()
        return runs_count
    except SignalReceivedException as sigex:
        # sleep this may raise SignalReceivedException
        logger.warning(f"Exiting runner due to signal {sigex.signal_code}")
        raise
    finally:
        if pool:
            logger.info("Closing processing pool")
            pool.shutdown(wait=True)
            pool = None
            logger.info("Processing pool closed")



from typing import Iterable, Iterator, Optional, List, Tuple
from pathlib import Path
from pathspec import PathSpec
from pathspec.util import iter_tree_files

from dlt._workspace._workspace_context import WorkspaceRunContext

# fallback ignore patterns used when no ignore file is found in the workspace
DEFAULT_IGNORES: List[str] = [
    "__pycache__/",
    "*.py[cod]",
    ".venv/",
    "venv/",
    "dist/",
    "build/",
    "*.egg-info/",
    ".mypy_cache/",
    ".ruff_cache/",
    ".pytest_cache/",
    ".coverage",
    "htmlcov/",
    "*.so",
    ".DS_Store",
    ".env",
]


class BaseFileSelector(Iterable[Tuple[Path, Path]]):
    """
    Base class for file selectors. For every file yields 2 paths: absolute path in the filesystem
    and relative path of the file in the resulting tarball
    """

    pass


class WorkspaceFileSelector(BaseFileSelector):
    """Iterates files in workspace respecting ignore patterns and excluding workspace internals.

    Uses gitignore-style patterns from a configurable ignore file (default .gitignore). Additional
    patterns can be provided as relative paths from workspace root. Settings directory is always excluded.
    """

    def __init__(
        self,
        context: WorkspaceRunContext,
        additional_excludes: Optional[List[str]] = None,
        ignore_file: str = ".gitignore",
    ) -> None:
        self.root_path: Path = Path(context.run_dir).resolve()
        self.settings_dir: Path = Path(context.settings_dir).resolve()
        self.ignore_file: str = ignore_file
        self.ignore_file_found: bool = False
        self.ignore_spec: PathSpec = self._build_pathspec(additional_excludes or [])

    def _build_pathspec(self, additional_excludes: List[str]) -> PathSpec:
        """Build PathSpec from ignore file + defaults + additional excludes"""
        patterns: List[str] = [f"{self.settings_dir.relative_to(self.root_path)}/"]

        # load ignore file if exists, otherwise fall back to default ignores
        ignore_path = self.root_path / self.ignore_file
        if ignore_path.exists():
            with ignore_path.open("r", encoding="utf-8") as f:
                patterns.extend(f.read().splitlines())
            self.ignore_file_found = True
        else:
            patterns.extend(DEFAULT_IGNORES)

        # Add caller-provided excludes
        patterns.extend(additional_excludes)

        return PathSpec.from_lines("gitwildmatch", patterns)

    def __iter__(self) -> Iterator[Tuple[Path, Path]]:
        """Yield paths of files eligible for deployment"""
        root_path = Path(self.root_path)
        for file_path in iter_tree_files(self.root_path):
            if not self.ignore_spec.match_file(file_path):
                yield root_path / file_path, Path(file_path)


class ConfigurationFileSelector(BaseFileSelector):
    """Iterates config and secrets files in workspace"""

    def __init__(
        self,
        context: WorkspaceRunContext,
    ) -> None:
        self.settings_dir: Path = Path(context.settings_dir).resolve()

    def __iter__(self) -> Iterator[Tuple[Path, Path]]:
        """Yield paths of config and secrets paths"""
        for file_path in iter_tree_files(self.settings_dir):
            if file_path.endswith("config.toml") or file_path.endswith("secrets.toml"):
                yield self.settings_dir / file_path, Path(file_path)



from typing import Union
from urllib.parse import urlparse, ParseResult

from dlt.common.exceptions import TypeErrorWithKnownTypes


def convert_storage_to_http_scheme(
    url: Union[str, ParseResult],
    use_https: bool = False,
    endpoint: str = None,
    region: str = None,
) -> str:
    try:
        if isinstance(url, str):
            parsed_url = urlparse(url)
        elif isinstance(url, ParseResult):
            parsed_url = url
        else:
            raise TypeErrorWithKnownTypes("url", url, ["str", "ParsedResult"])

        bucket_name = parsed_url.netloc
        object_key = parsed_url.path.lstrip("/")
        protocol = "https" if use_https else "http"

        if endpoint:
            domain = endpoint.replace("https://", "").replace("http://", "")
        elif region and parsed_url.scheme == "s3":
            domain = f"s3-{region}.amazonaws.com"
        else:
            storage_domains = {
                "s3": "s3.amazonaws.com",
                "gs": "storage.googleapis.com",
                "gcs": "storage.googleapis.com",
            }
            domain = storage_domains[parsed_url.scheme]

        return f"{protocol}://{bucket_name}.{domain}/{object_key}"
    except Exception as e:
        raise Exception(f"Error converting storage URL to HTTP protocol: `{url}`") from e



from typing import Any, Optional, Type, Union, Dict, TYPE_CHECKING


from dlt.common.destination import Destination, DestinationCapabilitiesContext
from dlt.common.normalizers.naming import NamingConvention

from dlt.destinations.impl.qdrant.configuration import QdrantCredentials, QdrantClientConfiguration

if TYPE_CHECKING:
    from dlt.destinations.impl.qdrant.qdrant_job_client import QdrantClient
else:
    QdrantClient = Any


class qdrant(Destination[QdrantClientConfiguration, "QdrantClient"]):
    spec = QdrantClientConfiguration

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        caps = DestinationCapabilitiesContext()
        caps.preferred_loader_file_format = "jsonl"
        caps.supported_loader_file_formats = ["jsonl"]
        caps.has_case_sensitive_identifiers = True
        caps.max_identifier_length = 200
        caps.max_column_identifier_length = 1024
        caps.max_query_length = 8 * 1024 * 1024
        caps.is_max_query_length_in_bytes = False
        caps.max_text_data_type_length = 8 * 1024 * 1024
        caps.is_max_text_data_type_length_in_bytes = False
        caps.supports_ddl_transactions = False
        caps.supported_replace_strategies = ["truncate-and-insert"]
        caps.supported_merge_strategies = ["upsert"]

        return caps

    @classmethod
    def adjust_capabilities(
        cls,
        caps: DestinationCapabilitiesContext,
        config: QdrantClientConfiguration,
        naming: Optional[NamingConvention],
    ) -> DestinationCapabilitiesContext:
        caps = super(qdrant, cls).adjust_capabilities(caps, config, naming)
        if config.is_local():
            # Local qdrant can not load in parallel
            caps.loader_parallelism_strategy = "sequential"
            caps.max_parallel_load_jobs = 1
        return caps

    @property
    def client_class(self) -> Type["QdrantClient"]:
        from dlt.destinations.impl.qdrant.qdrant_job_client import QdrantClient

        return QdrantClient

    def __init__(
        self,
        credentials: Union[QdrantClient, QdrantCredentials, Dict[str, Any]] = None,
        location: str = None,
        path: str = None,
        destination_name: str = None,
        environment: str = None,
        **kwargs: Any,
    ) -> None:
        """Configure the Qdrant destination to use in a pipeline.

        All arguments provided here supersede other configuration sources such as environment variables and dlt config files.

        Args:
            credentials (Union[QdrantClient, QdrantCredentials, Dict[str, Any]], optional): Credentials to connect to the Qdrant database. Can be an instance of `QdrantClient` or
                a dictionary with the credentials parameters.
            location (str, optional): The location of the Qdrant database.
            path (str, optional): The path to the Qdrant database.
            destination_name (str, optional): Name of the destination, can be used in config section to differentiate between multiple of the same type
            environment (str, optional): Environment of the destination
            **kwargs (Any, optional): Additional arguments passed to the destination config
        """
        super().__init__(
            credentials=credentials,
            location=location,
            path=path,
            destination_name=destination_name,
            environment=environment,
            **kwargs,
        )


qdrant.register()



from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.schema import Schema


from dlt.destinations.impl.duckdb.duck import DuckDbClient
from dlt.destinations.impl.motherduck.sql_client import MotherDuckSqlClient
from dlt.destinations.impl.motherduck.configuration import MotherDuckClientConfiguration


class MotherDuckClient(DuckDbClient):
    def __init__(
        self,
        schema: Schema,
        config: MotherDuckClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        dataset_name, staging_dataset_name = DuckDbClient.create_dataset_names(schema, config)
        super().__init__(schema, config, capabilities)  # type: ignore
        sql_client = MotherDuckSqlClient(
            dataset_name,
            staging_dataset_name,
            config.credentials,
            capabilities,
        )
        self.config: MotherDuckClientConfiguration = config  # type: ignore
        self.sql_client: MotherDuckSqlClient = sql_client
